from kuibit.timeseries import TimeSeries as kuibit_ts
import sys, os
sys.path.append(os.path.dirname(os.path.dirname(__file__)))
from gwBOB import gen_utils
import numpy as np
from gwBOB import BOB_utils
import sxs
import scri
import pytest

file_prefix = "./tests" #github
#file_prefix = "." #local
@pytest.fixture(scope="session")
def BOB_cce():
    wf_paths = {}
    wf_paths['h'] =  f'{file_prefix}/sxs_cache/cce9/rhOverM_BondiCce_R0270.h5'
    wf_paths['Psi4'] =  f'{file_prefix}/sxs_cache/cce9/rMPsi4_BondiCce_R0270.h5'
    wf_paths['Psi3'] =  f'{file_prefix}/sxs_cache/cce9/r2Psi3_BondiCce_R0270.h5'
    wf_paths['Psi2'] =  f'{file_prefix}/sxs_cache/cce9/r3Psi2OverM_BondiCce_R0270.h5'
    wf_paths['Psi1'] =  f'{file_prefix}/sxs_cache/cce9/r4Psi1OverM2_BondiCce_R0270.h5'
    wf_paths['Psi0'] =  f'{file_prefix}/sxs_cache/cce9/r5Psi0OverM3_BondiCce_R0270.h5'

    abd = scri.SpEC.create_abd_from_h5(file_format='RPDMB', **wf_paths)

    BOB = BOB_utils.BOB()
    BOB.initialize_with_cce_data(-1,provide_own_abd = abd,l=2,m=-2)

    return BOB
def BOB_params(initialize , location = f'{file_prefix}/trusted_outputs/BOB_BBH_2325_optimize_psi4.npz'):
    #Default to SXS BBH 2325 Params, Optimize Omega_0 = True if initialize and location are not given
    if initialize == None:
        data = np.load(location)
        params = ([data["mf"], data["chif"], data["l"], data["m"], data["Ap"], data["tp"], 
                   data["Omega_0"], data["Phi_0"], data["tau"], data["Omega_ISCO"]])
    elif initialize == "SXS":
        data = np.load(f'{file_prefix}/trusted_outputs/BOB_BBH_2325_optimize_psi4.npz')
        params = ([data["mf"], data["chif"], data["l"], data["m"], data["Ap"], data["tp"], 
                   data["Omega_0"], data["Phi_0"], data["tau"], data["Omega_ISCO"]])
    elif initialize == "CCE":
        data = np.load(f'{file_prefix}/trusted_outputs/BOB_BBH_CCE9_l2mm2_optimize_news.npz')
        params = ([data["mf"], data["chif"], data["l"], data["m"], data["Ap"], data["tp"], 
                   data["Omega_0"], data["Phi_0"], data["tau"], data["Omega_ISCO"]])
    return params
    
def kuibit_ts_load(location):
    data = np.load(location)
    timeseries = {}
    for key in data.files:
        if key.endswith("_t"):
            name = key[:-2]
            t = data[f"{name}_t"]
            y = data[f"{name}_y"]
            timeseries[name] = kuibit_ts(t, y)
    return timeseries

def test_initialize_with_sxs_data():
    # Set path for cache locally
    cache_path = f'{file_prefix}/sxs_cache'
    sxs.write_config(cache_directory=cache_path)

    expected_params = BOB_params("SXS")

    BOB = BOB_utils.BOB()
    BOB.initialize_with_sxs_data("SXS:BBH:2325",l=2,m=2,download=True)
    
    BOB.what_should_BOB_create = "psi4"
    BOB.optimize_Omega0 = True
    t_bob_psi4, y_bob_psi4 = BOB.construct_BOB()
    ts_psi4 = kuibit_ts(t_bob_psi4, y_bob_psi4)

    result_params = ([BOB.mf, BOB.chif, BOB.l, BOB.m, BOB.Ap, BOB.tp, 
           BOB.Omega_0, BOB.Phi_0, BOB.tau, BOB.Omega_ISCO])
    
    BOB.what_should_BOB_create = "news"
    BOB.optimize_Omega0 = True
    t_bob_news, y_bob_news = BOB.construct_BOB()
    ts_news = kuibit_ts(t_bob_news, y_bob_news)

    
    BOB.what_should_BOB_create = "strain"
    BOB.optimize_Omega0 = True
    t_bob_strain, y_bob_strain = BOB.construct_BOB()
    ts_strain = kuibit_ts(t_bob_strain, y_bob_strain)


    BOB_exp = kuibit_ts_load(f'{file_prefix}/trusted_outputs/BBH_2325_BOB_wf.npz')
    psi4_exp = BOB_exp["psi4"]
    news_exp = BOB_exp["news"]
    strain_exp = BOB_exp["strain"]

    mismatches = ([gen_utils.mismatch(ts_psi4, psi4_exp, t0 = 0, tf = 100), 
                   gen_utils.mismatch(ts_news, news_exp, t0 = 0, tf = 100), 
                   gen_utils.mismatch(ts_strain, strain_exp, t0 = 0, tf = 100)])
    mismatches_exp = ([0.0,0.0,0.0])

    for exp, res in zip(expected_params, result_params):
        assert np.isclose(exp, res, rtol=1e-12)
    for exp, res in zip(mismatches, mismatches_exp):
        assert np.isclose(exp, res, rtol=1e-12)

def test_initialize_with_cce_data(BOB_cce):


    expected_params = BOB_params(initialize = "CCE")

    BOB_cce.what_should_BOB_create = "strain"
    BOB_cce.optimize_Omega0 = True

    t,y = BOB_cce.construct_BOB()
    ts_strain = kuibit_ts(t,y)
    
    BOB_cce.what_should_BOB_create = "news"
    t,y = BOB_cce.construct_BOB()
    ts_news = kuibit_ts(t,y)
    result_params = ([BOB_cce.mf, BOB_cce.chif, BOB_cce.l, BOB_cce.m, BOB_cce.Ap, BOB_cce.tp, 
           BOB_cce.Omega_0, BOB_cce.Phi_0, BOB_cce.tau, BOB_cce.Omega_ISCO])
    
    BOB_cce.what_should_BOB_create = "psi4"
    t,y = BOB_cce.construct_BOB()
    ts_psi4 = kuibit_ts(t,y)

    BOB_exp = kuibit_ts_load(f'{file_prefix}/trusted_outputs/BBH_CCE9_l2mm2_BOB_wf.npz')
    print(BOB_exp)
    psi4_exp = BOB_exp["psi4"]
    news_exp = BOB_exp["news"]
    strain_exp = BOB_exp["strain"]

    mismatches = ([gen_utils.mismatch(ts_psi4, psi4_exp, t0 = 0, tf = 100), 
                   gen_utils.mismatch(ts_news, news_exp, t0 = 0, tf = 100), 
                   gen_utils.mismatch(ts_strain, strain_exp, t0 = 0, tf = 100)])
    mismatches_exp = ([0.0,0.0,0.0])

    for exp, res in zip(expected_params, result_params):
        assert np.isclose(exp, res, rtol=1e-12)
    for exp, res in zip(mismatches, mismatches_exp):
        assert np.isclose(exp, res, rtol=1e-12)
def test_kuibit_frequency_lm(BOB_cce):
    BOB_cce.what_should_BOB_create = "psi4"
    BOB_cce.optimize_Omega0 = True
    t,y = BOB_cce.construct_BOB()
    ts = kuibit_ts(t,y)
    freq = gen_utils.get_frequency(ts)
    # Load reference
    ref = np.load(f'{file_prefix}/trusted_outputs/kuibit_cce9_rMPsi4_R0270_freq_l2_mm2.npz')
    # Compare arrays
    np.testing.assert_allclose(freq.t, ref["f_t"], rtol=1e-10, atol=1e-15)
    np.testing.assert_allclose(freq.y, ref["f_y"], rtol=1e-10, atol=1e-15)

def test_get_phase(BOB_cce):
    BOB_cce.what_should_BOB_create = "psi4"
    BOB_cce.optimize_Omega0 = True
    t,y = BOB_cce.construct_BOB()
    ts = kuibit_ts(t,y)
    phase = gen_utils.get_phase(ts)
    # Load reference
    ref = np.load(f'{file_prefix}/trusted_outputs/kuibit_cce9_rMPsi4_R0270_phase_l2_mm2.npz')
    # Compare arrays
    np.testing.assert_allclose(phase.t, ref["phase_t"], rtol=1e-10, atol=1e-15)
    np.testing.assert_allclose(phase.y, ref["phase_y"], rtol=1e-10, atol=1e-15)


def test_get_r_isco_values():
    # Small arrays of chi and M
    chi_vals = np.array([0.0, 0.5, 0.9])
    M_vals = np.array([1.0, 2.0, 5.0])

    # Expected values computed manually or from reference
    expected = [
        6.0,  # (0, 1), Schwarzschild ISCO = 6M
        8.466005059061652, #(0.5, 2)
        11.604415208809435 # (0.9, 5.0)
    ]

    # Check that function returns correct shape and matches expected
    for chi, M, exp in zip(chi_vals, M_vals, expected):
        result = gen_utils.get_r_isco(chi, M)
        assert np.isclose(result, exp, rtol=1e-11)
def test_get_Omega_isco_values():
    chi_vals = np.array([0.0, 0.5, 0.9])
    M_vals = np.array([1.0, 2.0, 5.0])

    expected = [
        0.06804138174397717,
        0.05429417949013838,
        0.0450883417670616
    ]

    for chi, M, exp in zip(chi_vals, M_vals, expected):
        result = gen_utils.get_Omega_isco(chi, M)
        assert np.isclose(result, exp, rtol=1e-11)
def test_get_qnm():
    chi_vals = np.array([0.0, 0.0, 0.0, 0.5, 0.5, 0.5])
    M_vals = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 2.0])
    l_vals = np.array([2, 3, 2, 2, 2, 2]) 
    m_vals = np.array([2, 2, 2, 2, 2, 2]) 
    n_vals = np.array([0, 0, 1, 0, 0, 0])
    sign_vals = np.array([1, 1, 1, -1, 1, 1]) 

    
    expected_w_r_vals = np.array([0.37367168441804177, 0.5994432884374902, 0.34671099687916285, 0.32430731434882354, 0.46412302597593846, 0.23206151298796923])
    expected_tau_vals = np.array([11.24071459084527, 10.787131838360468, 3.6507692360145394, 11.231973996651769, 11.676945396785948, 23.353890793571896])


    for chi, M, l, m, n, sgn, exp_w, exp_tau in zip(chi_vals, M_vals, l_vals, m_vals, n_vals, sign_vals, expected_w_r_vals, expected_tau_vals):
        result_w, result_tau = gen_utils.get_qnm(chi, M, l, m, n = n, sign = sgn)
        assert np.isclose(result_w, exp_w, rtol=1e-11)
        assert np.isclose(result_tau, exp_tau, rtol=1e-11)
def test_get_tp_Ap_from_spline(BOB_cce):
    BOB_cce.what_should_BOB_create = "psi4"
    BOB_cce.optimize_Omega0 = True
    t,y = BOB_cce.construct_BOB()
    ts = kuibit_ts(t,y)
    amp = np.abs(ts)
    expected_tp, expected_Ap = ([5148.657477586399, 0.046735948589431364])
    result_tp, result_Ap = gen_utils.get_tp_Ap_from_spline(amp)
    assert np.isclose(result_tp, expected_tp, rtol=1e-11)
    assert np.isclose(result_Ap, expected_Ap, rtol=1e-11)



from jax import vmap, jit, checkpoint,jvp, config
import jax.numpy as jnp
from functools import partial

config.update("jax_enable_x64", True)


#NOTES: 
#1. THE OMEGA FUNCTIONS RETURN SMALL OMEGA, w=m*Omega.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!THIS IS DIFFERENT THAN IN BOB_terms.py!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#2. We don't implement phase functions here since the derivatives for the asymptotic expansion only involve A(t) and w(t)


#JAX versions

I = 1j

# Define the JAX-compatible BOB class
class JAXBOB:
    def __init__(self, t, Omega_0, Omega_QNM, tau, Ap, tp,m):
        self.t = jnp.array(t)
        self.Omega_0 = Omega_0
        self.Omega_QNM = Omega_QNM
        self.tau = tau
        self.Ap = Ap
        self.tp = tp
        self.m = abs(m)

def convert_BOB_to_JAXBOB(BOB):
    #t0_tp_tau = getattr(BOB, "t0_tp_tau", None) 
    #t0 = getattr(BOB, "t0", None)
    temp =  JAXBOB(BOB.t, BOB.Omega_0, BOB.Omega_QNM, BOB.tau, BOB.Ap,BOB.tp,BOB.m)
    return temp

def BOB_amplitude_jax(t, tau, Ap, t_p):
    '''
    BOB amplitude evolution

    Eq.5 in https://arxiv.org/abs/1810.00040

    Args:
        t : Time 
        tp : Time of peak amplitude
        tau : Damping time; inverse of the imaginary part of the QNM frequency
        Ap : Peak Waveform Amplitude

    Returns:
        A(t) : Waveform amplitude at time t
    '''
    tt = (t - t_p) / tau
    return Ap / jnp.cosh(tt)

def BOB_news_freq_jax(t, Omega_0, Omega_QNM, tau, t_p, m):
    '''
    Waveform frequency for the news when the BOB amplitude models the news (taking t_0 = -inf)

    Args:
        t : Time 
        t_p : Time of peak amplitude
        tau : Damping time; inverse of the imaginary part of the QNM frequency
        Omega_0 : Initial Condition Frequency
        Omega_QNM : Real part of Quasinormal mode (QNM) frequency/(mode number)
        m (int): Mode number

    Returns:
        omega - News waveform frequency
    '''
    tt = (t - t_p) / tau
    Omega_minus = Omega_QNM**2 - Omega_0**2
    Omega_plus  = Omega_QNM**2 + Omega_0**2
    Omega2 = Omega_minus * jnp.tanh(tt) / 2. + Omega_plus / 2.
    return m*jnp.sqrt(jnp.maximum(Omega2, 1e-12)) 

def BOB_news_phase_jax(t, Omega_0, Omega_QNM, tau, t_p, Phi_0, m=2):
    '''
    Waveform phase for the news when the BOB amplitude models the news (taking t_0 = -inf)

    Args:
        t : Time 
        t_p : Time of peak amplitude
        tau : Damping time; inverse of the imaginary part of the QNM frequency
        Omega_0 : Initial Condition Frequency
        Omega_QNM : Real part of Quasinormal mode (QNM) frequency/(mode number)
        Phi_0 : Initial Condition Phase (phi)/(mode number)
        m (int): Mode number

    Returns:
        phi - News waveform phase

        omega - News waveform frequency
    '''
    omega = BOB_news_freq_jax(t, Omega_0, Omega_QNM, tau, t_p, m) #news_freq_jax returns little omega
    Omega = omega/m

    Omega_minus_Q = jnp.abs(Omega - Omega_QNM) 
    Omega_minus_0 = jnp.abs(Omega - Omega_0)   
    
    # Handle the log(0) case safely by adding a small epsilon
    epsilon = 1e-40
    Omega_minus_Q = jnp.where(Omega_minus_Q == 0, epsilon, Omega_minus_Q)
    Omega_minus_0 = jnp.where(Omega_minus_0 == 0, epsilon, Omega_minus_0)


    outer = tau / 2.0
    inner1 = jnp.log(Omega + Omega_QNM) - jnp.log(Omega_minus_Q)
    inner2 = jnp.log(Omega + Omega_0) - jnp.log(Omega_minus_0)
    
    phase = (outer * (Omega_QNM * inner1 - Omega_0 * inner2) + Phi_0)*m
    
    return phase,omega
def BOB_psi4_freq_jax(t, Omega_0, Omega_QNM, tau, t_p,m):
    '''
    Waveform frequency for psi4 when assuming the BOB amplitude best models psi4 (taking t_0 = -inf)

    Args:
        t : Time 
        t_p : Time of peak amplitude
        tau : Damping time; inverse of the imaginary part of the QNM frequency
        Omega_0 : Initial Condition Frequency
        Omega_QNM : Real part of Quasinormal mode (QNM) frequency/(mode number)
        m (int): Mode number

    Returns:
        omega - Psi4 waveform frequency
    '''
    tt = (t - t_p) / tau
    k = (Omega_QNM**4 - Omega_0**4) / 2.0
    X = Omega_0**4 + k * (jnp.tanh(tt) + 1.0)
    return m*jnp.sqrt(jnp.sqrt(jnp.maximum(X, 1e-12)))

def BOB_strain_freq(t, Omega_0, Omega_QNM, tau, t_p,m):
    '''
    Waveform frequency for strain when assuming the BOB amplitude best models the strain (taking t_0 = -inf)

    Args:
        t : Time 
        t_p : Time of peak amplitude
        tau : Damping time; inverse of the imaginary part of the QNM frequency
        Omega_0 : Initial Condition Frequency
        Omega_QNM : Real part of Quasinormal mode (QNM) frequency/(mode number)
        m (int): Mode number

    Returns:
        omega - Strain waveform frequency
    '''
    tt = (t - t_p) / tau
    Omega_ratio = Omega_0/Omega_QNM
    tanh_tt_m1 = jnp.tanh(tt)-1
    return m*Omega_QNM*(Omega_ratio**(tanh_tt_m1/(-2.)))

def BOB_psi4_freq_finite_t0(t, Omega_0, Omega_QNM, tau, t_0, t_p,m):
    '''
    Waveform frequency for psi4 when assuming the BOB amplitude best models psi4 (for finite t_0)

    Args:
        t : Time 
        t_0 : Initial Condition time
        t_p : Time of peak amplitude
        tau : Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Omega_0 : Initial Condition Frequency
        Omega_QNM : Real part of Quasinormal mode (QNM) frequency/(mode number)
        m (int): Mode number

    Returns:
        omega - Psi4 waveform frequency
    '''
    tt = (t - t_p) / tau
    t0p = (t_0-t_p) / tau
    k_denom = 1 - jnp.tanh(t0p)
    k = (Omega_QNM**4 - Omega_0**4) / k_denom
    X = Omega_0**4 + k * (jnp.tanh(tt) - jnp.tanh(t0p))
    return m*(jnp.sqrt(jnp.sqrt(jnp.maximum(X, 1e-12))))

def BOB_news_freq_finite_t0(t, Omega_0, Omega_QNM, tau, t_0, t_p,m):
    '''
    Waveform frequency for the news when assuming the BOB amplitude best models the news (for finite t_0)

    Args:
        t : Time 
        t_0 : Initial Condition time
        t_p : Time of peak amplitude
        tau : Damping time; inverse of the imaginary part of the QNM frequency
        Omega_0 : Initial Condition Frequency
        Omega_QNM : Real part of Quasinormal mode (QNM) frequency/(mode number)
        m (int): Mode number

    Returns:
        omega - News waveform frequency
    '''
    tt = (t - t_p) / tau
    t0p = (t_0-t_p) / tau
    F_denom = 1 - jnp.tanh(t0p)
    F = (Omega_QNM**2 - Omega_0**2) / F_denom
    Omega2 = Omega_QNM**2 + F * (jnp.tanh(tt) - 1)
    return m*jnp.sqrt(jnp.maximum(Omega2, 1e-12))

def BOB_strain_freq_finite_t0(t, Omega_0, Omega_QNM, tau, t_0, t_p,m):
    '''
    Waveform frequency for the strain when assuming the BOB amplitude best models the strain (for finite t_0)
    
    Args:
        t : Time 
        t_0 : Initial Condition time
        t_p : Time of peak amplitude
        tau : Damping time; inverse of the imaginary part of the QNM frequency
        Omega_0 : Initial Condition Frequency
        Omega_QNM : Real part of Quasinormal mode (QNM) frequency/(mode number)
        m (int): Mode number

    Returns:
        omega - Strain waveform frequency  
    '''
    tt = (t - t_p) / tau
    t0p = (t_0-t_p) / tau
    Omega_ratio = Omega_0/Omega_QNM
    tanh_tt_m1 = jnp.tanh(tt)-1
    tanh_t0p_m1 = jnp.tanh(t0p)-1
    return m*Omega_QNM*(Omega_ratio**(tanh_tt_m1/tanh_t0p_m1))

def complex_scalar_derivative(g):
    """
    Compute the derivative of a complex scalar function g(t).
    """
    def deriv_g(t):
        # The Jacobian-vector product of g(t) with tangent vector 1.0 gives g'(t).
        _, g_prime = jvp(g, (t,), (1.0,))
        return g_prime
    return deriv_g
@partial(jit, static_argnames=('omega_func', 'A_func','N'))
def get_series_terms_ad(t, Omega_0, Omega_QNM, tau, Ap, t_p, omega_func, A_func, m, N):
    """
    Generates the raw, unsigned series terms [f₀, Df₀, D²f₀, ..., Dⁿf₀]
    using JAX's automatic differentiation for t_0 = -inf scenarios

    Args:
        t : Time 
        Omega_0 : Initial Condition Frequency
        Omega_QNM : Real part of Quasinormal mode (QNM) frequency/(mode number)
        tau : Damping time; inverse of the imaginary part of the QNM frequency
        Ap : Peak waveform amplitude 
        t_p : Time of peak amplitude
        omega_func: frequency function
        A_func: amplitude function  
        m : Mode number
        N : number of terms in the series

    Returns:
        A 2D array of shape (N+1, len(t)) containing the raw terms.
    """
    # Define the base function f₀(t) = A(t) / (i * ω(t))
    def f0_func(time):
        A = A_func(time, tau, Ap, t_p)
        omega = omega_func(time, Omega_0, Omega_QNM, tau, t_p, m)
        return A / (1j * omega)

    # Define the operator D's pre-factor g(t) = 1 / (i * ω(t))
    def g_func(time):
        omega = omega_func(time, Omega_0, Omega_QNM, tau, t_p, m)
        return 1.0 / (1j * omega)

    # List to hold the functions that compute [f₀, Df₀, D²f₀, ...]
    term_funcs = [f0_func]
    
    # Recursively build the derivative functions
    for i in range(1, N + 1):
        prev_term_func = term_funcs[-1]
        
        # Use jax.checkpoint to break the computational graph and save memory
        # during compilation for deep derivative chains.
        prev_term_func_checkpointed = checkpoint(prev_term_func)
        
        deriv_of_prev = complex_scalar_derivative(prev_term_func_checkpointed)
        
        next_term_func = lambda t, g=g_func, deriv=deriv_of_prev: g(t) * deriv(t)
        term_funcs.append(next_term_func)

    # Evaluate all functions over the time array using vmap
    all_terms = jnp.stack([vmap(f)(t) for f in term_funcs])
    
    return all_terms
@partial(jit, static_argnames=('omega_func', 'A_func','N'))
def get_series_terms_ad_finite_t0(t, Omega_0, Omega_QNM, tau, Ap, t_p, t_0, omega_func, A_func, m, N):
    """
    Generates the raw, unsigned series terms [f₀, Df₀, D²f₀, ..., Dⁿf₀]
    using JAX's automatic differentiation for finite t0 scenarios.

    Args:
        t : Time 
        Omega_0 : Initial Condition Frequency
        Omega_QNM : Real part of Quasinormal mode (QNM) frequency/(mode number)
        tau : Damping time; inverse of the imaginary part of the QNM frequency
        Ap : Peak waveform amplitude 
        t_p : Time of peak amplitude
        omega_func: frequency function
        A_func: amplitude function  
        m : Mode number
        N : number of terms in the series
    


    Returns:
        A 2D array of shape (N+1, len(t)) containing the raw terms.
    """
    # Define the base function f₀(t) = A(t) / (i * ω(t))
    def f0_func(time):
        A = A_func(time, tau, Ap, t_p)
        omega = omega_func(time, Omega_0, Omega_QNM, tau, t_0, t_p, m)
        return A / (1j * omega)

    # Define the operator D's pre-factor g(t) = 1 / (i * ω(t))
    def g_func(time):
        omega = omega_func(time, Omega_0, Omega_QNM, tau, t_0, t_p, m)
        return 1.0 / (1j * omega)

    # List to hold the functions that compute [f₀, Df₀, D²f₀, ...]
    term_funcs = [f0_func]
    
    # Recursively build the derivative functions
    for i in range(1, N + 1):
        prev_term_func = term_funcs[-1]
        
        # Use jax.checkpoint to break the computational graph and save memory
        # during compilation for deep derivative chains.
        prev_term_func_checkpointed = checkpoint(prev_term_func)
        
        deriv_of_prev = complex_scalar_derivative(prev_term_func_checkpointed)
        
        next_term_func = lambda t, g=g_func, deriv=deriv_of_prev: g(t) * deriv(t)
        term_funcs.append(next_term_func)

    # Evaluate all functions over the time array using vmap
    all_terms = jnp.stack([vmap(f)(t) for f in term_funcs])
    
    return all_terms

@partial(jit)
def fast_truncated_sum(all_raw_terms):
    """
    Calculates the series sum, including the alternating signs.
    
    Args:
        all_raw_terms: 2D array of shape (N+1, n_times) of UNSIGNED terms.

    Returns:
        Sum of all_raw_terms with alternating signs
    """
    N_plus_1 = all_raw_terms.shape[0]
    
    # Create the signs vector [1, -1, 1, -1, ...]
    signs = jnp.power(-1.0, jnp.arange(N_plus_1)).reshape(-1, 1)
    
    # Apply signs and sum along the terms axis (axis=0)
    series_sum = jnp.sum(all_raw_terms * signs, axis=0)
    
    return series_sum
@partial(jit, static_argnames=('omega_func', 'A_func','N'))
def calculate_strain_from_news(t, Omega_0, Omega_QNM, tau, Ap, t_p, 
                                 omega_func, A_func, m, N):
    '''
    Calculate the strain from the news using the series aapproximation
    Generates the raw, unsigned series terms [f₀, Df₀, D²f₀, ..., Dⁿf₀]
    using JAX's automatic differentiation for scenarios with t_0 = -inf

    Args:
        t : Time 
        Omega_0 : Initial Condition Frequency
        Omega_QNM : Real part of Quasinormal mode (QNM) frequency/(mode number)
        tau : Damping time; inverse of the imaginary part of the QNM frequency
        Ap : Peak waveform amplitude 
        t_p : Time of peak amplitude
        omega_func: frequency function
        A_func: amplitude function  
        m : Mode number
        N : number of terms in the series

    Returns:
        list of raw series terms (excluding alternating signs)

        series sum
    '''
    # 1. Generate the raw, unsigned derivative terms
    all_raw_terms = get_series_terms_ad(t, Omega_0, Omega_QNM, tau, Ap, t_p,
                                        omega_func, A_func, m, N)
    sum = fast_truncated_sum(all_raw_terms)
    return all_raw_terms,sum

@partial(jit, static_argnames=('omega_func', 'A_func','N'))
def calculate_strain_from_news_finite_t0(t, Omega_0, Omega_QNM, tau, Ap, t_p, t_0,
                                 omega_func, A_func, m, N):
    '''
    Calculate the strain from the news using the series aapproximation
    Generates the raw, unsigned series terms [f₀, Df₀, D²f₀, ..., Dⁿf₀]
    using JAX's automatic differentiation for scenarios invoving finite t_0 values.

    Args:
        t : Time 
        Omega_0 : Initial Condition Frequency
        Omega_QNM : Real part of Quasinormal mode (QNM) frequency/(mode number)
        tau : Damping time; inverse of the imaginary part of the QNM frequency
        Ap : Peak waveform amplitude 
        t_p : Time of peak amplitude
        omega_func: frequency function
        A_func: amplitude function  
        m : Mode number
        N : number of terms in the series

    Returns:
        list of raw series terms (excluding alternating signs)
        
        series sum
    '''
    # 1. Generate the raw, unsigned derivative terms
    all_raw_terms = get_series_terms_ad_finite_t0(t, Omega_0, Omega_QNM, tau, Ap, t_p, t_0,
                                        omega_func, A_func, m, N)
    sum = fast_truncated_sum(all_raw_terms)
    return all_raw_terms,sum

def _build_symbolic_series(base_func, g_func, N_order):
    """
    Internal helper to recursively build a list of symbolic derivative functions.
    
    Args:
        base_func: The starting function for the series (T₀).
        g_func: The function defining the pre-factor for the D operator.
        N_order: The truncation order for the series.
        
    Returns:
        A list of N+1 functions representing [T₀, T₁, ..., Tₙ].
    """
    term_funcs = [base_func]
    for _ in range(N_order):
        prev_func = term_funcs[-1]
        deriv_of_prev = complex_scalar_derivative(checkpoint(prev_func))
        # The core recursive definition of the D operator
        next_func = lambda time, g=g_func, d=deriv_of_prev: g(time) * d(time)
        term_funcs.append(next_func)
    return term_funcs

@partial(jit, static_argnames=('omega_func', 'A_func', 'N'))
def calculate_strain_from_psi4(t, Omega_0, Omega_QNM, tau, Ap, t_p,
                                            omega_func, A_func, m, N):
    '''
    Calculate the strain from psi4 using the series aapproximation
    Generates the raw, unsigned series terms [f₀, Df₀, D²f₀, ..., Dⁿf₀]
    using JAX's automatic differentiation for scenarios with t_0 = -inf.

    Args:
        t : Time 
        Omega_0 : Initial Condition Frequency
        Omega_QNM : Real part of Quasinormal mode (QNM) frequency/(mode number)
        tau : Damping time; inverse of the imaginary part of the QNM frequency
        Ap : Peak waveform amplitude 
        t_p : Time of peak amplitude
        omega_func: frequency function
        A_func: amplitude function  
        m : Mode number
        N : number of terms in both the inner and outer series

    Returns:
        list of raw series terms (excluding alternating signs)
        
        series sum
    '''
    M = N
    # --- Define the single D operator pre-factor ONCE ---
    def g_func(time):
        omega = omega_func(time, Omega_0, Omega_QNM, tau, t_p, m)
        return 1.0 / (1j * omega)

    # --- Stage 1: Build the symbolic function for the News sum ---
    def f0_psi4_func(time):
        A = A_func(time, tau, Ap, t_p)
        return A / (1j * omega_func(time, Omega_0, Omega_QNM, tau, t_p, m))

    news_series_term_funcs = _build_symbolic_series(f0_psi4_func, g_func, N)
    
    def news_sum_func(time):
        terms = jnp.stack([f(time) for f in news_series_term_funcs])
        signs = jnp.power(-1.0, jnp.arange(N + 1))
        return jnp.sum(terms * signs)

    # --- Stage 2: Build the symbolic function for the Strain sum ---
    def f0_strain_func(time):
        # The "amplitude" is the full sum from the previous stage
        A_news = news_sum_func(time)
        return A_news / (1j * omega_func(time, Omega_0, Omega_QNM, tau, t_p, m))

    strain_series_term_funcs = _build_symbolic_series(f0_strain_func, g_func, M)

    all_raw_terms_for_strain = jnp.stack(
        [vmap(f)(t) for f in strain_series_term_funcs]
    )

    # --- Stage 4: Sum the final terms ---
    strain_sum = fast_truncated_sum(all_raw_terms_for_strain)
    
    #return strain_sum

    return all_raw_terms_for_strain,strain_sum

@partial(jit, static_argnames=('omega_func', 'A_func', 'N'))
def calculate_strain_from_psi4_finite_t0(t, Omega_0, Omega_QNM, tau, Ap, t_p,
                                            t_0,omega_func, A_func, m, N):
    '''
    Calculate the strain from psi4 using the series aapproximation
    Generates the raw, unsigned series terms [f₀, Df₀, D²f₀, ..., Dⁿf₀]
    using JAX's automatic differentiation for scenarios with finite t_0.

    Args:
        t : Time 
        Omega_0 : Initial Condition Frequency
        Omega_QNM : Real part of Quasinormal mode (QNM) frequency/(mode number)
        tau : Damping time; inverse of the imaginary part of the QNM frequency
        Ap : Peak waveform amplitude 
        t_p : Time of peak amplitude
        omega_func: frequency function
        A_func: amplitude function  
        m : Mode number
        N : number of terms in both the inner and outer series

    Returns:
        list of raw series terms (excluding alternating signs)
        
        series sum
    '''
    M = N
    # --- Define the single D operator pre-factor ONCE ---
    def g_func(time):
        omega = omega_func(time, Omega_0, Omega_QNM, tau, t_0, t_p, m)
        return 1.0 / (1j * omega)

    # --- Stage 1: Build the symbolic function for the News sum ---
    def f0_psi4_func(time):
        A = A_func(time, tau, Ap, t_p)
        return A / (1j * omega_func(time, Omega_0, Omega_QNM, tau, t_0, t_p, m))

    news_series_term_funcs = _build_symbolic_series(f0_psi4_func, g_func, N)
    
    def news_sum_func(time):
        terms = jnp.stack([f(time) for f in news_series_term_funcs])
        signs = jnp.power(-1.0, jnp.arange(N + 1))
        return jnp.sum(terms * signs)

    # --- Stage 2: Build the symbolic function for the Strain sum ---
    def f0_strain_func(time):
        # The "amplitude" is the full sum from the previous stage
        A_news = news_sum_func(time)
        return A_news / (1j * omega_func(time, Omega_0, Omega_QNM, tau, t_0, t_p, m))

    strain_series_term_funcs = _build_symbolic_series(f0_strain_func, g_func, M)

    all_raw_terms_for_strain = jnp.stack(
        [vmap(f)(t) for f in strain_series_term_funcs]
    )

    # --- Stage 4: Sum the final terms ---
    strain_sum = fast_truncated_sum(all_raw_terms_for_strain)
    
    #return strain_sum

    return all_raw_terms_for_strain,strain_sum


import jax.numpy as jnp
import numpy as np
from jax import config

from gwBOB import BOB_terms_jax
from gwBOB import BOB_terms

config.update("jax_enable_x64", True)

def generate_strain_from_news_using_series(BOB,N=2):
    """
    Generate the strain from the news using the series approximation described in Kankani and McWilliams (2025) for t0 = -inf scenarios

    Parameters
    ----------
    BOB : BOB
        The BOB object containing the time series and parameters.
    N : int, optional
        The number of terms in both the inner and outer series. Default is 2.

    Returns
    -------
    t : numpy array
        The time array.
    h : numpy array
        The strain time series. The conjugate of the complex valued array is taken for consistency with SXS.
    """
    t = jnp.array(BOB.t)
    all_terms,sum = BOB_terms_jax.calculate_strain_from_news( t,
                                BOB.Omega_0,
                                BOB.Omega_QNM,
                                BOB.tau,
                                BOB.Ap,
                                BOB.tp,
                                BOB_terms_jax.BOB_news_freq_jax,
                                BOB_terms_jax.BOB_amplitude_jax,
                                BOB.m,
                                N)
    Phi,_ = BOB_terms.BOB_news_phase(BOB)
    phase_term = np.exp(1j * BOB.m * Phi)
    h = sum * phase_term  
    h = np.conj(h) #take conjugate for consistency with SXS
    return t,h

def generate_strain_from_psi4_using_series(BOB,N=2):
    """
    Generate the strain from psi4 using the series approximation described in Kankani and McWilliams (2025) for t0 = -inf scenarios.
    This is generally a suboptimal approach for psi4->strain but is included for testing.

    Parameters
    ----------
    BOB : BOB
        The BOB object containing the time series and parameters.
    N : int, optional
        The number of terms in both the inner and outer series. Default is 2.

    Returns
    -------
    t : numpy array
        The time array.
    h : numpy array
        The strain time series. The conjugate of the complex valued array is taken for consistency with SXS.
    """
    t = jnp.array(BOB.t)
    all_terms,sum = BOB_terms_jax.calculate_strain_from_psi4( t,
                                BOB.Omega_0,
                                BOB.Omega_QNM,
                                BOB.tau,
                                BOB.Ap,
                                BOB.tp,
                                BOB_terms_jax.BOB_psi4_freq_jax,
                                BOB_terms_jax.BOB_amplitude_jax,
                                BOB.m,
                                N)
    Phi,_ = BOB_terms.BOB_psi4_phase(BOB)
    phase_term = np.exp(1j * BOB.m * Phi)
    h = sum * phase_term  
    h = np.conj(h) #take conjugate for consistency with SXS
    return t,h


def generate_strain_from_news_using_series_finite_t0(BOB,N=2):
    """
    Generate the strain from the news using the series approximation described in Kankani and McWilliams (2025) for finite t0 scenarios.

    Parameters
    ----------
    BOB : BOB
        The BOB object containing the time series and parameters.
    N : int, optional
        The number of terms in both the inner and outer series. Default is 2.

    Returns
    -------
    t : numpy array
        The time array.
    h : numpy array
        The strain time series. The conjugate of the complex valued array is taken for consistency with SXS.
    """
    t = jnp.array(BOB.t)
    all_terms,sum = BOB_terms_jax.calculate_strain_from_news_finite_t0( t,
                                BOB.Omega_0,
                                BOB.Omega_QNM,
                                BOB.tau,
                                BOB.Ap,
                                BOB.tp,
                                BOB.t0,
                                BOB_terms_jax.BOB_news_freq_finite_t0,
                                BOB_terms_jax.BOB_amplitude_jax,
                                BOB.m,
                                N)
    Phi,_ = BOB_terms.BOB_news_phase_finite_t0(BOB)
    phase_term = np.exp(1j * BOB.m * Phi)
    h = sum * phase_term  
    h = np.conj(h) #take conjugate for consistency with SXS
    return t,h

def generate_strain_from_psi4_using_series_finite_t0(BOB,N=2):
    """
    Generate the strain from the psi4 using the series approximation described in Kankani and McWilliams (2025) for finite t0 scenarios.
    This is generally a suboptimal approach for psi4->strain but is included for testing.

    Parameters
    ----------
    BOB : BOB
        The BOB object containing the time series and parameters.
    N : int, optional
        The number of terms in both the inner and outer series. Default is 2.

    Returns
    -------
    t : numpy array
        The time array.
    h : numpy array
        The strain time series. The conjugate of the complex valued array is taken for consistency with SXS.
    """
    t = jnp.array(BOB.t)
    all_terms,sum = BOB_terms_jax.calculate_strain_from_psi4_finite_t0( t,
                                BOB.Omega_0,
                                BOB.Omega_QNM,
                                BOB.tau,
                                BOB.Ap,
                                BOB.tp,
                                BOB.t0,
                                BOB_terms_jax.BOB_psi4_freq_finite_t0,
                                BOB_terms_jax.BOB_amplitude_jax,
                                BOB.m,
                                N)
    Phi,_ = BOB_terms.BOB_psi4_phase_finite_t0(BOB)
    phase_term = np.exp(1j * BOB.m * Phi)
    h = sum * phase_term  
    h = np.conj(h) #take conjugate for consistency with SXS
    return t,h




from . import BOB_utils
from . import gen_utils
from . import convert_to_strain_using_series
from . import mismatch_utils
from . import BOB_terms
from . import BOB_terms_jax
from . import ascii_funcs

__all__ = ["BOB_utils", "gen_utils", "convert_to_strain_using_series", "mismatch_utils", "BOB_terms", "BOB_terms_jax", "ascii_funcs"]
__version__ = "0.0.1"


#pyright: reportUnreachable=false
#JAX implemented mismatch search

#Notes:
#This code is only meant for merger-ringdown searches. It performs a time domain integration since we want to calculate the mismatch over a fixed time window.

#When we pass in EOB/surrogate data, we align at peak with the NR data beforehand, so the ideal time shift should be close to 0

#Instead of a grid based phase search, we find the best phase by maximizing the overlap
#https://journals.aps.org/prd/pdf/10.1103/PhysRevD.85.122006 section 4, text after eq 4.1


from functools import partial
from jax import jit,vmap
import jax.numpy as jnp
#uncomment if we want to use cubic spline integration
#import interpax
from jax import debug
#jax.config.update("jax_log_compiles", True)

@partial(jit)
def time_shift(h_complex, t, t_shift):
    """
    JAX compatible time shift function. Shifts the time series by a given amount.

    Parameters
    ----------
    h_complex : complex array
        The complex-valued time series to be shifted.
    t : array
        The time array.
    t_shift : float
        The amount to shift the time series by.

    Returns
    -------
    h_shifted : complex array
        The shifted time series.
    """
    shifted_time_grid = t - t_shift
    h_shifted_real = jnp.interp(shifted_time_grid, t, h_complex.real)
    h_shifted_imag = jnp.interp(shifted_time_grid, t, h_complex.imag)
    return h_shifted_real + 1j * h_shifted_imag

@partial(jit, static_argnames=('integration_points',))
def mismatch_trapz(
    h1_padded, t1_padded, # Original, unshifted Model data
    h2_padded, t2_padded, # NR data
    t_peak_nr,
    t0_relative, tf_relative,
    integration_points
):
    """
    JAX compatible mismatch function. Calculates the mismatch between two time series using the trapz integration method.

    Parameters
    ----------
    h1_padded : complex array
        The complex-valued time series of the model data.
    t1_padded : array
        The time array of the model data.
    h2_padded : complex array
        The complex-valued time series of the NR data.
    t2_padded : array
        The time array of the NR data.
    t_peak_nr : float
        The peak time of the NR data.
    t0_relative : float
        The relative start time of the integration window.
    tf_relative : float
        The relative end time of the integration window.
    integration_points : int
        The number of integration points.

    Returns
    -------
    mismatch : float
        The mismatch between the two time series.
    """
    
    t_start_abs = t_peak_nr + t0_relative
    t_end_abs = t_peak_nr + tf_relative
    t_integ = jnp.linspace(t_start_abs, t_end_abs, integration_points)

    h1_integ = jnp.interp(t_integ, t1_padded, h1_padded.real, left=0.0, right=0.0) + \
          1j * jnp.interp(t_integ, t1_padded, h1_padded.imag, left=0.0, right=0.0)

    h2_integ = jnp.interp(t_integ, t2_padded, h2_padded.real, left=0.0, right=0.0) + \
          1j * jnp.interp(t_integ, t2_padded, h2_padded.imag, left=0.0, right=0.0)
 
    # Numerator
    numerator_integrand = jnp.conj(h1_integ) * h2_integ
    numerator_integral = jnp.trapezoid(numerator_integrand, t_integ)
    
    # Denominators
    denom1_integrand = jnp.real(jnp.conj(h1_integ) * h1_integ)
    denom2_integrand = jnp.real(jnp.conj(h2_integ) * h2_integ)
    denom1_sq = jnp.trapezoid(denom1_integrand, t_integ)
    denom2_sq = jnp.trapezoid(denom2_integrand, t_integ)
    
    denominator1 = jnp.sqrt(denom1_sq)
    denominator2 = jnp.sqrt(denom2_sq)
    
    epsilon = 1e-20
    maximized_overlap = jnp.abs(numerator_integral) / (denominator1 * denominator2 + epsilon)
    best_phi0 = -jnp.angle(numerator_integral)
    mismatch = 1.0 - maximized_overlap
    
    return mismatch

#uncomment if we want to use cubic spline integration
# @partial(jit, static_argnames=('integration_points'))
# def mismatch_interpax(
#     h1_padded, t1_padded,      #model data  
#     h2_padded, t2_padded,      #nr data          
#     t_peak_nr,               
#     t0_relative, tf_relative,integration_points):


#     t_start_abs = t_peak_nr + t0_relative
#     t_end_abs = t_peak_nr + tf_relative

#     t_integ = jnp.linspace(t_start_abs,t_end_abs,integration_points)

#     # Resample h2 onto t1's grid.
#     # left=0.0, right=0.0 ensures padded regions outside t2's domain become zero.
#     h1_common = jnp.interp(t_integ,t1_padded,h1_padded, left=0.0, right=0.0)
#     h2_common = jnp.interp(t_integ, t2_padded, h2_padded, left=0.0, right=0.0)
#     h1_integ = jnp.interp(t_integ, t1_padded, h1_padded.real, left=0.0, right=0.0) + \
#           1j * jnp.interp(t_integ, t1_padded, h1_padded.imag, left=0.0, right=0.0)

#     h2_integ = jnp.interp(t_integ, t2_padded, h2_padded.real, left=0.0, right=0.0) + \
#           1j * jnp.interp(t_integ, t2_padded, h2_padded.imag, left=0.0, right=0.0)

    
#     numerator_integrand = jnp.conj(h1_integ) * h2_integ
#     denom1_integrand = jnp.real(jnp.conj(h1_integ) * h1_integ)
#     denom2_integrand = jnp.real(jnp.conj(h2_integ) * h2_integ)
    
#     numerator_integral = interpax.CubicSpline(
#         x = t_integ, 
#         y = numerator_integrand,
#         check=False
#     ).integrate(t_start_abs, t_end_abs)

#     denom1_sq = interpax.CubicSpline(
#         x = t_integ, 
#         y = denom1_integrand,
#         check=False
#     ).integrate(t_start_abs, t_end_abs)
    
#     denom2_sq = interpax.CubicSpline(
#         x = t_integ, 
#         y = denom2_integrand,
#         check=False
#     ).integrate(t_start_abs, t_end_abs)

#     denominator1 = jnp.sqrt(jnp.real(denom1_sq))
#     denominator2 = jnp.sqrt(jnp.real(denom2_sq))
    
#     epsilon = 1e-20
#     #we take the absolute value of numerator_integral because that corresponds to the maximum overlap/ideal phase shift
#     maximized_overlap = jnp.abs(numerator_integral) / (denominator1 * denominator2 + epsilon)
#     #best_phi0 = -jnp.angle(numerator_integral)
    
#     mismatch = 1.0 - maximized_overlap
#     return mismatch

@partial(jit, static_argnames=('t0', 'tf', 'coarse_window', 'coarse_t_num', 'fine_window', 'fine_t_num','integration_points'))
def find_best_mismatch_padded(
    padded_t_model, padded_h_model,
    padded_t_nr, padded_h_nr,
    nr_peak_time_batch,
    t0, tf, coarse_window, coarse_t_num, fine_window, fine_t_num, integration_points
):
    '''
    JAX compatible mismatch search function. Finds the best time shift between two time series using the trapz integration method.
    
    Parameters
    ----------
    padded_t_model : array
        The time array of the model data.
    padded_h_model : complex array
        The complex-valued time series of the model data.
    padded_t_nr : array
        The time array of the NR data.
    padded_h_nr : complex array
        The complex-valued time series of the NR data.
    nr_peak_time_batch : array
        The peak time of the NR data.
    t0 : float
        The relative start time of the integration window.
    tf : float
        The relative end time of the integration window.
    coarse_window : float
        The coarse window size.
    coarse_t_num : int
        The number of coarse integration points.
    fine_window : float
        The fine window size.
    fine_t_num : int
        The number of fine integration points.
    integration_points : int
        The number of integration points.
    '''
    
    def find_best_for_one_waveform(t_m, h_m, t_n, h_n, nr_peak):

        t_range_1 = jnp.linspace(-coarse_window, coarse_window, coarse_t_num)
        
        @vmap
        def do_search(t_shift):
            h_m_shifted = time_shift(h_m, t_m, t_shift)
            return mismatch_trapz(
                h_m_shifted, t_m, h_n, t_n, nr_peak, t0, tf,integration_points
            )

        mismatches_1 = do_search(t_range_1)
        min_idx_1 = jnp.argmin(mismatches_1)
        
        mismatch_1 = mismatches_1[min_idx_1]
        t_shift_1 = t_range_1[min_idx_1]


        t_range_2 = jnp.linspace(
            t_shift_1 - fine_window, 
            t_shift_1 + fine_window, 
            fine_t_num 
        )
        
        mismatches_2 = do_search(t_range_2)
        min_idx_2 = jnp.argmin(mismatches_2)

        mismatch_2 = mismatches_2[min_idx_2]
        t_shift_2 = t_range_2[min_idx_2]
        
        #debug.print("t_shift_1 = {x}",x=t_shift_1)
        #debug.print("t_shift_2 = {x}",x=t_shift_2)

        is_fine_search_better = mismatch_2 < mismatch_1
        
        final_mismatch = jnp.where(is_fine_search_better, mismatch_2, mismatch_1)
        #final_t_shift = jnp.where(is_fine_search_better, t_shift_2, t_shift_1)
        #debug.print("final_t_shift = {x}",x=final_t_shift)
        return final_mismatch

    # --- Apply the entire 2-stage search to the batch of waveforms ---
    return vmap(find_best_for_one_waveform)(
        padded_t_model, padded_h_model,
        padded_t_nr, padded_h_nr,
        nr_peak_time_batch
    )


import numpy as np
from kuibit.timeseries import TimeSeries as kuibit_ts
import qnm
from quaternion.calculus import spline_definite_integral as sdi
import matplotlib.pyplot as plt
import scri
import spherical_functions as sf
from scipy.signal import butter, filtfilt, detrend, lfilter
from scipy.optimize import minimize, differential_evolution
from numpy import trapz
from scipy.interpolate import CubicSpline
import sxs



#some useful functions
def find_nearest_index(array, value):
    '''
    Return the index of the element in ``array`` closest to ``value``.

    Parameters
    ----------
    array : numpy.ndarray
        1D array to search.
    value : float
        Target value to find the nearest element to.

    Returns
    -------
    int
        Index of the nearest value in ``array``.
    '''
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx
def get_kuibit_lm(w,l,m):
    '''
    Extract the ``(l, m)`` mode from a ``sxs.WaveformModes`` and return a kuibit ``TimeSeries``.

    Parameters
    ----------
    w : sxs.WaveformModes
        Waveform containing spherical-harmonic modes.
    l : int
        Spherical-harmonic index ``l``.
    m : int
        Spherical-harmonic index ``m``.

    Returns
    -------
    kuibit_ts
        TimeSeries of the selected mode values over time.
    '''
    index = w.index(l, m)
    w_temp = w.data[:,index]
    time = w.t
    return kuibit_ts(time,w_temp)
def get_kuibit_lm_psi4(w,l,m):
    '''
    Extract the ``(l, m)`` mode from a ``sxs.WaveformModes`` psi4 object into a kuibit ``TimeSeries``.

    Notes
    -----
    This accessor expects a ``psi4``-like object where mode data is accessed as ``w[:, index].ndarray``.

    Parameters
    ----------
    w : sxs.WaveformModes
        Psi4 waveform containing spherical-harmonic modes.
    l : int
        Spherical-harmonic index ``l``.
    m : int
        Spherical-harmonic index ``m``.

    Returns
    -------
    kuibit_ts
        TimeSeries of the selected psi4 mode values over time.
    '''
    index = w.index(l, m)
    w_temp = w[:,index].ndarray
    time = w.t
    return kuibit_ts(time,w_temp)
def get_kuibit_frequency_lm(w,l,m):
    '''
    Compute the instantaneous angular frequency of the ``(l, m)`` mode as a kuibit ``TimeSeries``.

    The phase of the mode is unwrapped and differentiated to obtain angular frequency. The sign is
    flipped to ensure positive frequencies near merger.

    Parameters
    ----------
    w : scri.WaveformModes
        Waveform containing spherical-harmonic modes (psi4 recommended).
    l : int
        Spherical-harmonic index ``l``.
    m : int
        Spherical-harmonic index ``m``.

    Returns
    -------
    kuibit_ts
        TimeSeries of angular frequency for the selected mode.
    '''
    ts = get_kuibit_lm_psi4(w,l,m)
    #returns the time derivative of np.unwrap(np.angle(w.y))
    ts_temp = ts.phase_angular_velocity()
    #want positive
    return kuibit_ts(ts_temp.t,-ts_temp.y)
def get_phase(ts):
    '''
    Return the unwrapped phase of a complex ``TimeSeries``.

    Parameters
    ----------
    ts : Kuibit TimeSeries
        Complex-valued time series.

    Returns
    -------
    Kuibit TimeSeries
        TimeSeries of unwrapped phase. Sign may be flipped to be positive near merger.
    '''
    y = np.unwrap(np.angle(ts.y))
    #we want to make sure the phase is positive near merger so this how we check for now
    #TODO: make this better
    if(y[-1]<0):
        y = -y
    return kuibit_ts(ts.t,y)
def get_frequency(ts):
    '''
    Compute the instantaneous angular frequency of a complex ``TimeSeries``.

    Parameters
    ----------
    ts : Kuibit TimeSeries
        Complex-valued time series.

    Returns
    -------
    Kuibit TimeSeries
        TimeSeries of angular frequency. Sign is chosen positive near peak amplitude.
    '''
    tp = ts.time_at_maximum()
    freq = ts.phase_angular_velocity()
    if(freq.y[find_nearest_index(freq.t,tp)]<0):
        freq.y = -freq.y
    return kuibit_ts(ts.t,freq.y)
def get_r_isco(chi_with_sign,M):
    '''
    Compute the prograde ISCO radius for a Kerr black hole. For negative chi_with_sign values, we return the retrograde value
    
    Parameters
    ----------
    chi_with_sign : float
        Dimensionless spin of the remnant. Negative value indicates a spin pointed opposite the direction of the initial anular momentum of the binary.
    M : float
        Mass of the remnant (in geometric units).

    Returns
    -------
    float
        ISCO radius.
    '''
    #Bardeen Press Teukolskly eq 2.21
    #defined for prograde orbits
    if(chi_with_sign>=0):
        sign = 1.0
    else:
        sign = -1.0
    chi = np.abs(chi_with_sign)
    a = chi*M
    a_M = a/M

    z1 = 1 + (((1-a_M**2)**(1./3.)) * ((1+a_M)**(1./3.) + (1-a_M)**(1./3.))) #good
    z2 = (3*(a_M**2) + z1**2)**0.5 #good
    r_isco = M * (3 + z2 - (sign)*((3-z1)*(3+z1+2*z2))**0.5) #good
    return r_isco
def get_Omega_isco(chi_with_sign,M):
    '''
    Compute the orbital angular velocity at the ISCO for a Kerr black hole.
    For negative chi_with_sign values, we return the retrograde value

    Parameters
    ----------
    chi_with_sign : float
        Dimensionless spin of the remnant. Negative value indicates a spin pointed opposite the direction of the initial anular momentum of the binary.
    M : float
        Mass of the remnant (in geometric units).

    Returns
    -------
    float
        ISCO angular velocity.
    '''
    #Bardeen Press Teukolskly eq 2.16
    #defined for prograde orbits
    if(chi_with_sign>=0):
        sign = 1.0
    else:
        sign = -1.0
    chi = np.abs(chi_with_sign)
    r_isco = get_r_isco(chi_with_sign,M)
    a = chi*M
    Omega = sign*np.sqrt(M)/(r_isco**1.5 + sign*a*np.sqrt(M)) # = dphi/dt
    return Omega

def get_qnm(chif,Mf,l,m,n=0,sign=1):
    '''
    Get the fundamental quasinormal mode frequency components for a Kerr black hole.

    Parameters
    ----------
    chif : float
        Dimensionless final spin magnitude (chi).
    Mf : float
        Final black-hole mass (geometric units).
    l : int
        Spherical-harmonic index ``l``.
    m : int
        Spherical-harmonic index ``m`` (sign handled via ``sign``).
    n : int, optional
        Overtone number, by default 0.
    sign : int, optional
        Mode sign (+1 for m, -1 for -m), by default 1.

    Returns
    -------
    tuple[float, float]
        ``(w_r, tau)`` where ``w_r`` is the real angular frequency and ``tau`` is the damping time.
    '''
    #omega_qnm, all_C, ells = qnmfits.read_qnms.qnm_from_tuple((l,m,n,1),chif,M=M)
    if(sign==-1):
        grav_lmn = qnm.modes_cache(s=-2,l=l,m=-m,n=n)
        omega_qnm, A, C = grav_lmn(a=chif)#qnm package uses M = 1 so a = chi here
        omega_qnm = -np.conj(omega_qnm)
    else:
        grav_lmn = qnm.modes_cache(s=-2,l=l,m=m,n=n)
        omega_qnm, A, C = grav_lmn(a=chif)#qnm package uses M = 1 so a = chi here
    omega_qnm /= Mf #rescale to remnant black hole mass
    w_r = np.abs(omega_qnm.real)
    imag_qnm = np.abs(omega_qnm.imag)
    tau = 1./imag_qnm
    return w_r,tau
def get_tp_Ap_from_spline(amp):
    '''
    Find the peak time and amplitude using a cubic-spline interpolation.

    Parameters
    ----------
    amp : Kuibit TimeSeries
        TimeSeries of amplitude (e.g., ``|h|`` or ``|psi4|``).

    Returns
    -------
    tuple[float, float]
        ``(tp, Ap)`` where ``tp`` is the peak time and ``Ap`` is the peak amplitude.
    '''
    #we assume junk radiation has been removed, so the largest amplitude is the physical peak
    spline = CubicSpline(amp.t,amp.y,extrapolate=False)
    dspline = spline.derivative()
    critical_points = dspline.roots()
    critical_points = critical_points[
        (critical_points >= amp.t[0]) & (critical_points <= amp.t[-1])
    ]
    y_candidates = spline(critical_points)
    max_idx = np.argmax(y_candidates)
    tp = critical_points[max_idx]
    Ap = y_candidates[max_idx]
    return tp,Ap
def mismatch(model_data,NR_data,t0,tf,use_trapz=False,resample_NR_to_model=True,return_best_phi0=False):   
    '''
    Compute the normalized mismatch between a model and reference time series.

    Parameters
    ----------
    model_data : Kuibit TimeSeries
        Model complex time series.
    NR_data : Kuibit TimeSeries
        Reference complex time series.
    t0 : float
        Start time relative to the reference peak time.
    tf : float
        End time relative to the reference peak time.
    use_trapz : bool, optional
        If True, use trapezoidal integration; otherwise use spline definite integral, by default False.
    resample_NR_to_model : bool, optional
        If True, resample NR data onto the model time grid when they differ, by default True.
    return_best_phi0 : bool, optional
        If True, also return the phase shift that maximizes the overlap, by default False.

    Returns
    -------
    float or tuple[float, float]
        Mismatch in [0, 1]. If ``return_best_phi0`` is True, also returns ``best_phi0``.
    '''
    #simple mismatch function
    if (not(np.array_equal(model_data.t,NR_data.t))):
        if(resample_NR_to_model):
            #print("resampling to equal times")
            NR_data = NR_data.resampled(model_data.t)
        else:
            raise ValueError("Time arrays must be identical or set resample_NR_to_model to True")
    
    peak_time = NR_data.time_at_maximum()
    
    
    dx = model_data.t[1] - model_data.t[0]

    if(use_trapz):
        NR_data = NR_data.cropped(init=peak_time+t0,end=peak_time+tf)
        model_data = model_data.cropped(init=peak_time+t0,end=peak_time+tf)

    numerator_integrand = np.conj(model_data.y)*NR_data.y
    if(use_trapz is False):
        numerator = (sdi(numerator_integrand,model_data.t,peak_time+t0,peak_time+tf))
    else:
        numerator = (trapz(numerator_integrand,model_data.t))
    
    denominator1_integrand = np.conj(model_data.y)*model_data.y
    if(use_trapz is False):
        denominator1 = np.real(sdi(denominator1_integrand,model_data.t,peak_time+t0,peak_time+tf))
    else:
        denominator1 = np.real(trapz(denominator1_integrand,model_data.t))
    
    denominator2_integrand = np.conj(NR_data.y)*NR_data.y
    if(use_trapz is False):
        denominator2 = np.real(sdi(denominator2_integrand,NR_data.t,peak_time+t0,peak_time+tf))
    else:
        denominator2 = np.real(trapz(denominator2_integrand,NR_data.t))
    
    #maximized overlap when numerator = |numerator|
    max_mismatch = (np.abs(numerator)/np.sqrt(denominator1*denominator2))
    best_phi0 = -np.angle(numerator)
    if(return_best_phi0):
        return 1.-max_mismatch,best_phi0
    return 1.-max_mismatch   
def time_grid_mismatch(model, NR_data, t0, tf, resample_NR_to_model=True,
                           t_shift_range=None,return_best_t_and_phi0=False):
    '''
    Search over time shifts to minimize mismatch between model and reference.

    Parameters
    ----------
    model : Kuibit TimeSeries
        Model complex time series.
    NR_data : Kuibit TimeSeries
        Reference complex time series.
    t0 : float
        Start time relative to reference peak.
    tf : float
        End time relative to reference peak.
    resample_NR_to_model : bool, optional
        If True, resample NR data onto model time grid, by default True.
    t_shift_range : numpy.ndarray, optional
        Range of time shifts to search, by default ``np.arange(-10, 10, 0.1)``.
    return_best_t_and_phi0 : bool, optional
        If True, also return the best time shift and phase shift, by default False.

    Returns
    -------
    float or tuple
        Minimum mismatch, and optionally best ``t_shift`` and ``phi0``.
    '''
    if(t_shift_range is None):
        t_shift_range = np.arange(-10,10,0.1)
    min_mismatch = np.inf
    def mismatch_search(t_shift_range,min_mismatch):
        '''
        Helper to scan a range of time shifts and track the best values.

        Parameters
        ----------
        t_shift_range : numpy.ndarray
            Candidate time shifts to test.
        min_mismatch : float
            Current best mismatch (updated in place).

        Returns
        -------
        tuple
            Either ``(min_mismatch, best_t_shift)`` or ``(min_mismatch, best_t_shift, best_phi0)``.
        '''

        best_t_shift = 0
        best_phi0 = 0 
        for t_shift in t_shift_range:
            model_ = kuibit_ts(model.t + t_shift,model.y)
            if(return_best_t_and_phi0):
                mismatch_val,phi0 = mismatch(model_,NR_data,t0,tf,use_trapz=True,resample_NR_to_model=resample_NR_to_model,return_best_phi0=True)
            else:
                mismatch_val = mismatch(model_,NR_data,t0,tf,use_trapz=True,resample_NR_to_model=resample_NR_to_model)
            if mismatch_val < min_mismatch:
                min_mismatch = mismatch_val
                best_t_shift = t_shift
                if(return_best_t_and_phi0):
                    best_phi0 = phi0

        if(return_best_t_and_phi0):
            return min_mismatch,best_t_shift,best_phi0
        return min_mismatch,best_t_shift
    
    if(return_best_t_and_phi0):
        min_mismatch,best_t_shift,best_phi0 = mismatch_search(t_shift_range,min_mismatch)
    else:
        min_mismatch,best_t_shift = mismatch_search(t_shift_range,min_mismatch)

    t_shift_range = np.arange(best_t_shift-0.2,best_t_shift+0.2,0.01)
    if(return_best_t_and_phi0):
        min_mismatch,best_t_shift,best_phi0 = mismatch_search(t_shift_range,min_mismatch)
        return min_mismatch,best_t_shift,best_phi0
    else:
        min_mismatch,best_t_shift = mismatch_search(t_shift_range,min_mismatch)
        return min_mismatch
def estimate_parameters(BOB,
                        mf_guess=0.95,
                        chif_guess=0.5,
                        Omega0_guess=0.155,
                        t0=0,
                        tf=75,
                        force_Omega0_optimization=False,
                        NR_data=None,
                        make_current_naturally=False,
                        make_mass_naturally=False,
                        include_Omega0_as_parameter=False,
                        include_2Omega0_as_parameters=False,
                        perform_phase_alignment_first=False,
                        start_with_wide_search = False,
                        t_shift_range=None):
    '''
    Estimate BOB parameters by minimizing mismatch against NR data.

    Parameters
    ----------
    BOB : object
        A configured BOB instance.
    mf_guess : float, optional
        Initial guess for remnant mass, by default 0.95.
    chif_guess : float, optional
        Initial guess for signed dimensionless spin, by default 0.5.
    Omega0_guess : float, optional
        Initial guess for initial condition frequency, by default 0.155.
    t0 : float, optional
        Start time after peak for mismatch window, by default 0.
    tf : float, optional
        End time after peak for mismatch window, by default 75.
    force_Omega0_optimization : bool, optional
        If True, enforce Omega0 optimization during construction, by default False.
    NR_data : kuibit_ts, optional
        Reference data to compare against. If None, inferred from ``BOB`` selection.
    make_current_naturally : bool, optional
        If True, construct current quadrupole via natural method, by default False.
    make_mass_naturally : bool, optional
        If True, construct mass quadrupole via natural method, by default False.
    include_Omega0_as_parameter : bool, optional
        Include Omega0 as an optimization parameter, by default False.
    include_2Omega0_as_parameters : bool, optional
        Include both lm and lmm Omega0 as parameters (quadrupole builds), by default False.
    perform_phase_alignment_first : bool, optional
        If True, perform phase alignment before quadrupole combination, by default False.
    start_with_wide_search : bool, optional
        If True, run a global search before local refinement, by default False.
    t_shift_range : numpy.ndarray, optional
        Range of time shifts to search in mismatch, by default ``np.arange(-10, 10, 0.1)``.

    Returns
    -------
    scipy.optimize.OptimizeResult
        Optimizer result containing best-fit parameters.
    '''
    if(t_shift_range is None):
        t_shift_range = np.arange(-10,10,0.1)
    if(force_Omega0_optimization and include_Omega0_as_parameter):
        raise ValueError("force_Omega0_optimization and include_Omega0_as_parameter cannot both be True")
    if(make_current_naturally is True and make_mass_naturally is True):
        raise ValueError("make_current_naturally and make_mass_naturally cannot both be True")
    if((force_Omega0_optimization and include_2Omega0_as_parameters) or (force_Omega0_optimization and include_Omega0_as_parameter)):
        raise ValueError("force_Omega0_optimization and include_2Omega0_as_parameters cannot both be True")
    if(include_2Omega0_as_parameters is True and include_Omega0_as_parameter is False):
        raise ValueError("include_2Omega0_as_parameters is True and include_Omega0_as_parameter is False")
    #store BOB parameters
    old_mf = BOB.mf
    old_chif = BOB.chif
    old_chif_with_sign = BOB.chif_with_sign
    old_Omega0 = BOB.Omega_0
    #we use a scipy optimizer to find the best mass and spin
    if(BOB.what_should_BOB_create=="psi4"):
        #Psi4
        A = 1.42968337
        B = 0.08424419
        C = -1.22848524
        NR_ts = BOB.psi4_data
    if(BOB.what_should_BOB_create=="news"):
        #News
        A = 0.33568227
        B = 0.03450997
        C = -0.18763176  
        NR_ts = BOB.news_data
        
    if(NR_data is not None):
        NR_ts = NR_data
    
    def create_guess(x):   
        '''
        Objective function: construct BOB for candidate parameters and return mismatch.

        Parameters
        ----------
        x : numpy.ndarray
            Parameter vector with order depending on options.

        Returns
        -------
        float
            Mismatch value (lower is better).
        '''
        #print("trying",x)
        mf = x[0]
        chif = x[1]
        if(include_Omega0_as_parameter):
            lm_Omega0_guess = x[2]
        if(include_2Omega0_as_parameters):
            lmm_Omega0_guess = x[3]
        BOB.fit_failed = False
        BOB.mf = mf
        BOB.chif_with_sign = chif
        BOB.chif = np.abs(chif)
        if(force_Omega0_optimization):
            BOB.optimize_Omega0 = True
            BOB.start_fit_before_tpeak = t0
            BOB.end_fit_after_tpeak = tf
        else:
            BOB.optimize_Omega0 = False
            BOB.Omega_0 = A*BOB.mf + B*BOB.chif_with_sign + C 

        if(include_Omega0_as_parameter):
            #keep this for ordinary (l,m) &(l,-m) modes
            BOB.Omega_0 = lm_Omega0_guess
        w_r,tau = get_qnm(BOB.chif,BOB.mf,BOB.l,np.abs(BOB.m),sign=np.sign(BOB.chif_with_sign))
        BOB.Omega_QNM = w_r/np.abs(BOB.m)
        BOB.Phi_0 = 0
        BOB.tau = tau
        BOB.t_tp_tau = (BOB.t - BOB.tp)/BOB.tau
        try:
            if(make_current_naturally is False and make_mass_naturally is False):
                t,y = BOB.construct_BOB()
            elif(make_current_naturally):
                if(include_2Omega0_as_parameters):
                    t,y = BOB.construct_BOB_current_quadrupole_naturally(perform_phase_alignment_first=perform_phase_alignment_first,lm_Omega0=lm_Omega0_guess,lmm_Omega0=lmm_Omega0_guess)
                elif(include_Omega0_as_parameter):
                    t,y = BOB.construct_BOB_current_quadrupole_naturally(perform_phase_alignment_first=perform_phase_alignment_first,lm_Omega0=lm_Omega0_guess)
                else:
                    t,y = BOB.construct_BOB_current_quadrupole_naturally(perform_phase_alignment_first=perform_phase_alignment_first)
            elif(make_mass_naturally):
                if(include_2Omega0_as_parameters):
                    t,y = BOB.construct_BOB_mass_quadrupole_naturally(perform_phase_alignment_first=perform_phase_alignment_first,lm_Omega0=lm_Omega0_guess,lmm_Omega0=lmm_Omega0_guess)
                elif(include_Omega0_as_parameter):  
                    t,y = BOB.construct_BOB_mass_quadrupole_naturally(perform_phase_alignment_first=perform_phase_alignment_first,lm_Omega0=lm_Omega0_guess)
                else:
                    t,y = BOB.construct_BOB_mass_quadrupole_naturally(perform_phase_alignment_first=perform_phase_alignment_first)
            else:
                raise ValueError("Invalid options for make_current_naturally and make_mass_naturally")
            BOB_ts = kuibit_ts(t,y)
            if(BOB.fit_failed):
                print("fit failed for ",x)
                mismatch = np.inf
            else:
                #print("fit worked for ",x)
                mismatch = time_grid_mismatch(BOB_ts,NR_ts,t0,tf,t_shift_range=t_shift_range)
        except Exception as e:
            mismatch = np.inf
            print(e)
            print("Search failed for ",x)
        return mismatch
    #we use nelder-mead because the mismatch can return infinity, causing problems with derivatives
    if(include_2Omega0_as_parameters):
        if(start_with_wide_search):
            out = differential_evolution(create_guess,bounds = [(0.8, 0.999), (-0.999,0.999), (0+1e-10,BOB.Omega_QNM-1e-10),(0+1e-10,BOB.Omega_QNM-1e-10)])
            out = minimize(create_guess,out.x,bounds = [(0.8, 0.999), (-0.999,0.999), (0+1e-10,BOB.Omega_QNM-1e-10),(0+1e-10,BOB.Omega_QNM-1e-10)],method='Nelder-Mead')
        else:
            out = minimize(create_guess,(mf_guess,chif_guess,Omega0_guess,Omega0_guess),bounds = [(0.8, 0.999), (-0.999,0.999), (0+1e-10,BOB.Omega_QNM-1e-10),(0+1e-10,BOB.Omega_QNM-1e-10)],method='Nelder-Mead')
    elif(include_Omega0_as_parameter):
        if(start_with_wide_search):
            out = differential_evolution(create_guess,bounds = [(0.8, 0.999), (-0.999,0.999), (0+1e-10,BOB.Omega_QNM-1e-10)])
            out = minimize(create_guess,out.x,bounds = [(0.8, 0.999), (-0.999,0.999), (0+1e-10,BOB.Omega_QNM-1e-10)],method='Nelder-Mead')
        else:
            out = minimize(create_guess,(mf_guess,chif_guess,Omega0_guess),bounds = [(0.8, 0.999), (-0.999,0.999), (0+1e-10,BOB.Omega_QNM-1e-10)],method='Nelder-Mead')
    else:
        if(start_with_wide_search):
            out = differential_evolution(create_guess,bounds = [(0.8, 0.999), (-0.999,0.999)])
            out = minimize(create_guess,out.x,bounds = [(0.8, 0.999), (-0.999,0.999)],method='Nelder-Mead')
        else:
            out = minimize(create_guess,(mf_guess,chif_guess),bounds = [(0.8, 0.999), (-0.999,0.999)],method='Nelder-Mead')
    #reset parameters in BOB
    BOB.mf = old_mf
    BOB.chif = old_chif
    BOB.chif_with_sign = old_chif_with_sign
    BOB.Omega_0 = old_Omega0
    return out

def create_QNM_comparison(t,y,NR_data,mov_time,tf,mf,chif,n_qnms=7):
    '''
    Compare a model against NR by fitting QNM sums over a moving end-time window.

    Parameters
    ----------
    t : numpy.ndarray
        Sampling times of the model.
    y : numpy.ndarray
        Complex model values at ``t``.
    NR_data : scri.WaveformModes
        Reference waveform mode data (scri object).
    mov_time : array-like
        Start times for the moving-window comparison (relative to peak after alignment).
    tf : float
        End time for the mismatch calculation window.
    mf : float
        Remnant mass.
    chif : float
        Signed dimensionless remnant spin.
    n_qnms : int, optional
        Maximum number of overtones to include in fits, by default 7.

    Returns
    -------
    tuple
        ``(master_mismatch_arr, A220_dict, A221_dict, A222_dict, qnm_wm_master_arr)``.
    '''
    import qnmfits
    #we use qnmfits for their qnm fitting procedure
    #TODO: use varpro instead
    #TODO: beyond (2,2) mode

    #mov_time is the time for the moving mismatch
    #tf is the time the mismatch is calculated until

    #NR_data must be a scri waveform mode.
    #I'm not dealing with the headache of converting t,y arrays to waveform modes

    #code based on https://github.com/sxs-collaboration/qnmfits/blob/main/examples/working_with_cce.ipynb

    model = kuibit_ts(t,y)
    t_peak = model.time_at_maximum()
    mov_time = mov_time + t_peak
    tf = tf + t_peak
    
    qnm_list = [[(2,2,n,1) for n in range(N)] for N in range(1,n_qnms+2)]
    spherical_modes = [(2,2)]

    A220_dict = {}
    A221_dict = {}
    A222_dict = {}
    master_mismatch_arr = []
    qnm_wm_master_arr = []
    for N,qnms in enumerate(qnm_list):
        A220_dict[N] = []
        A221_dict[N] = []
        A222_dict[N] = []
        mm_list = []
        qnm_wm_arr = []
        for start_time in mov_time:
            best_fit = qnmfits.fit(
                data=NR_data,
                chif=chif,
                Mf=mf,
                qnms=qnms,
                spherical_modes=spherical_modes,
                t0=start_time,
                T = tf-start_time #T is the duration of the mismatch calculation. We want to always end the mismatch calculation at tf
            )
            mm_list.append(best_fit['mismatch'])
            A220_dict[N].append(abs(best_fit['amplitudes'][2,2,0,1]))
            if(N>0):
                A221_dict[N].append(abs(best_fit['amplitudes'][2,2,1,1]))
            else:
                A221_dict[N].append(0)
            if(N>1):
                A222_dict[N].append(abs(best_fit['amplitudes'][2,2,2,1]))
            else:
                A222_dict[N].append(0)
            qnm_wm_arr.append(best_fit['model'])
        master_mismatch_arr.append(mm_list)
        qnm_wm_master_arr.append(qnm_wm_arr)
    return master_mismatch_arr,A220_dict,A221_dict,A222_dict,qnm_wm_master_arr
def create_scri_news_waveform_mode(times,y_22_data,ell_min=2,ell_max=None):
    '''
    Create a ``scri.WaveformModes`` object (news) from time samples and a single mode.

    Parameters
    ----------
    times : numpy.ndarray
        Sampling times.
    y_22_data : numpy.ndarray
        Complex data for the ``(2, 2)`` mode.
    ell_min : int, optional
        Minimum ``ell`` included in the WaveformModes, by default 2.
    ell_max : int, optional
        Maximum ``ell`` included. If None, inferred from provided data, by default None.

    Returns
    -------
    scri.WaveformModes
        Constructed news WaveformModes object.
    '''
    #based on https://github.com/sxs-collaboration/qnmfits/blob/main/qnmfits/utils.py  dict_to_WaveformModes
    #but modified for our purposes here

    #for now we only include the (2,2) mode
    data = {(2,2):y_22_data}
    if ell_max is None:
        ell_max = max([ell for ell, _ in data.keys()])

    # The spherical-harmonic mode (ell, m) indices for the requested ell_min
    # and ell_max
    ell_m_list = sf.LM_range(ell_min, ell_max)

    # Initialize the WaveformModes data array
    wm_data = np.zeros((len(times), len(ell_m_list)), dtype=complex)

    # Fill the WaveformModes data array
    for i, (ell, m) in enumerate(ell_m_list):
        if (ell, m) in data.keys():
            wm_data[:, i] = data[(ell, m)]

    # Construct the WaveformModes object
    wm = scri.WaveformModes(
        dataType=scri.news,
        t=times,
        data=wm_data,
        ell_min=ell_min,
        ell_max=ell_max,
        frameType=scri.Inertial,
        r_is_scaled_out=True,
        m_is_scaled_out=True
    )

    return wm
def create_scri_psi4_waveform_mode(times,y_22_data,ell_min=2,ell_max=None):
    '''
    Create a ``scri.WaveformModes`` object (psi4) from time samples and a single mode.

    Parameters
    ----------
    times : numpy.ndarray
        Sampling times.
    y_22_data : numpy.ndarray
        Complex data for the ``(2, 2)`` mode.
    ell_min : int, optional
        Minimum ``ell`` included in the WaveformModes, by default 2.
    ell_max : int, optional
        Maximum ``ell`` included. If None, inferred from provided data, by default None.

    Returns
    -------
    scri.WaveformModes
        Constructed psi4 WaveformModes object.
    '''
    #based on https://github.com/sxs-collaboration/qnmfits/blob/main/qnmfits/utils.py  dict_to_WaveformModes
    #but modified for our purposes here

    data = {(2,2):y_22_data}

    if ell_max is None:
        ell_max = max([ell for ell, _ in data.keys()])

    # The spherical-harmonic mode (ell, m) indices for the requested ell_min
    # and ell_max
    ell_m_list = sf.LM_range(ell_min, ell_max)

    # Initialize the WaveformModes data array
    wm_data = np.zeros((len(times), len(ell_m_list)), dtype=complex)

    # Fill the WaveformModes data array
    for i, (ell, m) in enumerate(ell_m_list):
        if (ell, m) in data.keys():
            wm_data[:, i] = data[(ell, m)]

    # Construct the WaveformModes object
    wm = scri.WaveformModes(
        dataType=scri.psi4,
        t=times,
        data=wm_data,
        ell_min=ell_min,
        ell_max=ell_max,
        frameType=scri.Inertial,
        r_is_scaled_out=True,
        m_is_scaled_out=True
    )

    return wm
def weighted_detrend(signal, weight_power=2):
    '''
    Perform weighted linear detrending with heavier weight on later samples.

    Parameters
    ----------
    signal : numpy.ndarray
        Real-valued signal to detrend.
    weight_power : float, optional
        Exponent controlling the late-time weighting, by default 2.

    Returns
    -------
    numpy.ndarray
        Detrended signal.
    '''
    n = len(signal)
    x = np.arange(n)

    # Emphasize later times more
    weights = (x / n) ** weight_power  # adjust power to control how sharp the weighting is

    # Fit weighted linear trend
    A = np.vstack([x, np.ones(n)]).T
    W = np.diag(weights)
    coeffs = np.linalg.lstsq(W @ A, W @ signal, rcond=None)[0]

    trend = A @ coeffs
    return signal - trend
def time_integral(ts,order=2,f=0.1,dt=0.1,remove_drift = False):
    '''
    Integrate a complex time series with high-pass filtering to control low-frequency drift.

    The signal is optionally resampled to fixed timestep, high-pass filtered, and cumulatively
    integrated. An optional weighted-detrend step can remove residual linear drift.

    Parameters
    ----------
    ts : kuibit_ts
        Complex time series to integrate.
    order : int, optional
        Butterworth filter order, by default 2.
    f : float, optional
        High-pass cutoff as a fraction of peak frequency, by default 0.1.
    dt : float, optional
        Desired timestep for resampling and integration, by default 0.1.
    remove_drift : bool, optional
        If True, apply weighted detrending to integral, by default False.

    Returns
    -------
    kuibit_ts
        Integrated complex time series.
    '''
    #time integral with a butterworth highpass filter and a digital filter to ensure the phase doesn't change
    #optional linear drift removal at end, with the highpass filter, it doesn't make much of a difference
    #Note: The phase after integration may not be the mismatch minimized phase for the new waveform, but should be pretty good
    if(np.abs((ts.t[-1]-ts.t[0])-dt)>1e-10):
        ts = ts.fixed_timestep_resampled(dt)
    freq = get_frequency(ts)
    peak_time = ts.time_at_maximum()
    freq_at_peak = freq.y[find_nearest_index(freq.t,peak_time)]/(2*np.pi)
    #assert(w_qnm/(2*np.pi)>freq_at_peak)
    fs = 1/dt
    b,a = butter(order,freq_at_peak*f/(.5*fs),btype='highpass',analog=False)
    #b,a = butter(order,[freq_at_peak*f/(.5*fs),(w_qnm*2.5/(2*np.pi))/(.5*fs)],btype='band',analog=False)
    filtered_signal_real = filtfilt(b, a, ts.y.real)
    filtered_signal_imag = filtfilt(b, a, ts.y.imag)
    if(remove_drift):
        real_int = weighted_detrend(np.cumsum(filtered_signal_real)/fs)
        imag_int = weighted_detrend(np.cumsum(filtered_signal_imag)/fs)
    else:
        real_int = np.cumsum(filtered_signal_real)/fs
        imag_int = np.cumsum(filtered_signal_imag)/fs
    return kuibit_ts(ts.t,real_int + 1j*imag_int)
def compute_one_more_term(nth_derivative,t,freq):
    '''
    Compute an additional correction term using spline differentiation and 1/(i omega).

    Parameters
    ----------
    nth_derivative : numpy.ndarray
        Values of the nth derivative.
    t : numpy.ndarray
        Time samples corresponding to ``nth_derivative``.
    freq : numpy.ndarray
        Angular frequency samples at ``t``.

    Returns
    -------
    numpy.ndarray
        Correction term values evaluated on ``t``.
    '''
    #we want to compute one final term on top of the autodifferentiated result
    one_over_iomega = 1/(-1j*freq)
    deriv_val = kuibit_ts(t,nth_derivative).spline_differentiated(1).y*one_over_iomega
    return deriv_val
def load_lower_lev_SXS(sim):
    '''
    Load the next-lower available resolution for an SXS simulation.

    Parameters
    ----------
    sim : sxs.Simulation
        The loaded SXS Simulation at some refinement level.

    Returns
    -------
    sxs.Simulation
        The lower-resolution Simulation, if available.

    Raises
    ------
    ValueError
        If only one level exists or the lower level cannot be found.
    '''
    location = sim.location
    print(location,sim.lev_numbers)
    if(len(sim.lev_numbers)>1):
       try:        
           sim_lower = sxs.load(location[:-1]+str(sim.lev_numbers[-2]))
       except:
            raise ValueError("Lower level not found")
    else:
        raise ValueError("only one Level found")
    return sim_lower
def Omega_0_fit_psi4(Mf,chif_with_sign):
    '''
    Omega0 for psi4 using the fit from Kankani and McWilliams (2025)
    
    Parameters
    ----------
    Mf : float
        Remnant mass.
    chif_with_sign : float
        Remnant spin. (negative values indicate a final spin pointing opposite to the initial orbital angular momentum)

    Returns
    -------
    float
        Omega_0 fit.
    
    '''
    A = 1.42968337
    B = 0.08424419
    C = -1.22848524
    return A*Mf + B*chif_with_sign + C
def Omega_0_fit_news(Mf,chif_with_sign):
    '''
    Omega0 for news using the fit from Kankani and McWilliams (2025)

    Parameters
    ----------
    Mf : float
        Remnant mass.
    chif_with_sign : float
        Remnant spin. (negative values indicate a final spin pointing opposite to the initial orbital angular momentum)

    Returns
    -------
    float
        Omega_0 fit.
    
    '''
    A = 0.33568227
    B = 0.03450997
    C = -0.18763176
    return A*Mf + B*chif_with_sign + C
def Omega_0_fit_strain(Mf,chif_with_sign):
    '''
    Omega0 for strain using the fit from Kankani and McWilliams (2025)

    Parameters
    ----------
    Mf : float
        Remnant mass.
    chif_with_sign : float
        Remnant spin. (negative values indicate a final spin pointing opposite to the initial orbital angular momentum)

    Returns
    -------
    float
        Omega_0 fit.
    
    '''
    A = 0.01663248
    B = 0.01798275
    C = 0.07882578
    return A*Mf + B*chif_with_sign + C
    




    







import numpy as np
from scipy.special import expi as Ei_np
from scipy.integrate import cumulative_trapezoid
import sympy as sp
from sympy import Ei as Ei_sp

def define_BOB_symbols():
    t, t0, tp, tau = sp.symbols('t t0 tp tau', real=True)
    Omega0, Omega_QNM, Ap, Phi_0 = sp.symbols('Omega0 Omega_QNM Ap Phi_0', positive=True)
    return t, t0, tp, tau, Omega0, Omega_QNM, Ap, Phi_0

def BOB_amplitude_sym(t, tp, tau, Ap):
    '''
    BOB amplitude evolution

    Eq.5 in https://arxiv.org/abs/1810.00040

    Args:
        t (sympy.Symbol): Time 
        tp (sympy.Symbol): Time of peak amplitude
        tau (sympy.Symbol): Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Ap (sympy.Symbol): Peak Waveform Amplitude

    Returns:
        A(t) - Waveform amplitude
    '''
    x = (t - tp) / tau
    return Ap * sp.sech(x)

def BOB_strain_freq_finite_t0_sym(t, t0, tp, tau, Omega0, Omega_QNM):
    '''
    Frequency for the strain when assuming the BOB amplitude best models the strain (for finite t_0)

    Args:
        t (sympy.Symbol): Time 
        t0 (sympy.Symbol): Initial Condition time
        tp (sympy.Symbol): Time of peak amplitude 
        tau (sympy.Symbol): Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Omega0 (sympy.Symbol): Initial Condition Frequency
        Omega_QNM (sympy.Symbol): Real part of Quasinormal mode (QNM) frequency (little omega)/(mode number)

    Returns:
        Omega -  Strain frequency [(waveform frequency)/(mode number)]
    '''
    x = (t - tp) / tau
    x0 = (t0 - tp) / tau
    Omega_ratio = Omega0 / Omega_QNM
    exponent = (sp.tanh(x) - 1) / (sp.tanh(x0) - 1)
    return Omega_QNM * Omega_ratio**exponent

def BOB_strain_phase_finite_t0_sym(t, t0, tp, tau, Omega0, Omega_QNM, Phi_0):
    '''
    Phase for the strain when the BOB amplitude models the strain (for finite t_0)

    Args:
        t (sympy.Symbol): Time 
        t0 (sympy.Symbol): Initial Condition time
        tp (sympy.Symbol): Time of peak amplitude
        tau (sympy.Symbol): Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Omega0 (sympy.Symbol): Initial Condition Frequency
        Omega_QNM (sympy.Symbol): Real part of Quasinormal mode (QNM) frequency (little omega)/(mode number)
        Phi_0 (sympy.Symbol): Initial Condition Phase (phi)/(mode number)

    Returns:
        Phi - Phase of strain (waveform phase)/(mode number)

        Omega - Strain frequency (waveform frequency)/(mode number)
    '''
    Omega = BOB_strain_freq_finite_t0_sym(t, t0, tp, tau, Omega0, Omega_QNM)
    x = (t - tp) / tau
    x0 = (t0 - tp) / tau
    Omega_ratio = Omega0 / Omega_QNM
    outer = Omega_QNM * tau / sp.Integer(2)
    tanh_tp_t0_tau_p1 = sp.tanh(-x0) + 1
    term1_exp = sp.Integer(2) / tanh_tp_t0_tau_p1
    term1 = Omega_ratio**term1_exp
    log_Omega_ratio_div_tanh = -sp.log(Omega_ratio) / (sp.tanh(x0) - 1)
    term2 = log_Omega_ratio_div_tanh * (sp.tanh(x) + 1)
    term3 = log_Omega_ratio_div_tanh * (sp.tanh(x) - 1)
    inner = term1 * Ei_sp(term2) - Ei_sp(term3)
    Phi = outer * inner + Phi_0
    return Phi, Omega

def BOB_news_freq_finite_t0_sym(t, t0, tp, tau, Omega0, Omega_QNM):
    '''
    Frequency for the news when the BOB amplitude models the news (for finite t_0)

    Args:
        t (sympy.Symbol): Time 
        t0 (sympy.Symbol): Initial Condition time
        tp (sympy.Symbol): Time of peak amplitude
        tau (sympy.Symbol): Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Omega0 (sympy.Symbol): Initial Condition Frequency
        Omega_QNM (sympy.Symbol): Real part of Quasinormal mode (QNM) frequency (little omega)/(mode number)

    Returns:

        Omega - News frequency
        
    '''
    x = (t - tp) / tau
    x0 = (t0 - tp) / tau
    F = (Omega_QNM**2 - Omega0**2) / (1 - sp.tanh(x0))
    Omega2 = Omega_QNM**2 + F * (sp.tanh(x) - 1)
    return sp.sqrt(Omega2)

def BOB_news_phase_finite_t0_sym(t, t0, tp, tau, Omega0, Omega_QNM, Phi_0):
    '''
    Phase for the news when the BOB amplitude models the News (for finite t_0)

    Args:
        t (sympy.Symbol): Time 
        t0 (sympy.Symbol): Initial Condition time
        tp (sympy.Symbol): Time of peak amplitude
        tau (sympy.Symbol): Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Omega0 (sympy.Symbol): Initial Condition Frequency
        Omega_QNM (sympy.Symbol): Real part of Quasinormal mode (QNM) frequency (little omega)/(mode number)
        Phi_0 (sympy.Symbol): Initial Condition Phase (phi)/(mode number)

    Returns:
        Phi - News phase (waveform phase)/(mode number)

        Omega - News frequency (waveform frequency)/(mode number)
    '''
    Omega = BOB_news_freq_finite_t0_sym(t, t0, tp, tau, Omega0, Omega_QNM)
    x0 = (t0 - tp) / tau
    F = (Omega_QNM**2 - Omega0**2) / (1 - sp.tanh(x0))
    delta = sp.Integer(2) * F - Omega_QNM**2
    term1 = Omega_QNM * tau / sp.Integer(2) * sp.log((Omega + Omega_QNM) / sp.Abs(Omega - Omega_QNM))
    term2 = -sp.sqrt(delta) * tau * sp.atan(Omega / sp.sqrt(delta))
    return term1 + term2 + Phi_0, Omega

def BOB_psi4_freq_finite_t0_sym(t, t0, tp, tau, Omega0, Omega_QNM):
    '''
    Frequency for Psi4 when the BOB amplitude models Psi4 (for finite t_0)

    Args:
        t (sympy.Symbol): Time 
        t0 (sympy.Symbol): Initial Condition time
        tp (sympy.Symbol): Time of peak amplitude
        tau (sympy.Symbol): Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Omega0 (sympy.Symbol): Initial Condition Frequency
        Omega_QNM (sympy.Symbol): Real part of Quasinormal mode (QNM) frequency (little omega)/(mode number)

    Returns:
        Omega - Psi4 frequency (waveform frequency)/(mode number)
    '''
    x = (t - tp) / tau
    x0 = (t0 - tp) / tau
    k = (Omega_QNM**4 - Omega0**4) / (1 - sp.tanh(x0))
    X = Omega0**4 + k * (sp.tanh(x) - sp.tanh(x0))
    return X**sp.Rational(1, 4)

def BOB_psi4_phase_finite_t0_sym(t, t0, tp, tau, Omega0, Omega_QNM, Phi_0):
    '''
    Phase for Psi4 when the BOB amplitude models Psi4 (for finite t_0)

    Args:
        t (sympy.Symbol): Time 
        t0 (sympy.Symbol): Initial Condition time
        tp (sympy.Symbol): Time of peak amplitude
        tau (sympy.Symbol): Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Omega0 (sympy.Symbol): Initial Condition Frequency
        Omega_QNM (sympy.Symbol): Real part of Quasinormal mode (QNM) frequency (little omega)/(mode number)
        Phi_0 (sympy.Symbol): Initial Condition Phase (phi)/(mode number)

    Returns:
        Phi - Psi4 phase (waveform phase)/(mode number)

        Omega - Psi4 frequency (waveform frequency)/(mode number)
    '''
    Omega = BOB_psi4_freq_finite_t0_sym(t, t0, tp, tau, Omega0, Omega_QNM)
    x0 = (t0 - tp) / tau
    k = (Omega_QNM**4 - Omega0**4) / (1 - sp.tanh(x0))
    KappaP = (Omega0**4 + k * (1 - sp.tanh(x0)))**sp.Rational(1, 4)
    KappaM = (Omega0**4 - k * (1 + sp.tanh(x0)))**sp.Rational(1, 4)
    term_P_tanh = KappaP * tau * (sp.atanh(Omega / KappaP) - sp.atanh(Omega0 / KappaP))
    term_M_tanh = KappaM * tau * (sp.atanh(Omega / KappaM) - sp.atanh(Omega0 / KappaM))
    term_P_tan  = KappaP * tau * (sp.atan(Omega / KappaP) - sp.atan(Omega0 / KappaP))
    term_M_tan  = KappaM * tau * (sp.atan(Omega / KappaM) - sp.atan(Omega0 / KappaM))
    return term_P_tanh + term_P_tan - term_M_tanh - term_M_tan + Phi_0, Omega

# --- Asymptotic (t0 -> -inf) ---
def BOB_strain_freq_sym(t, tp, Omega0, Omega_QNM, tau):
    '''
    Frequency for the strain when the BOB amplitude models the strain (for t_0 = -inf)

    Args:
        t (sympy.Symbol): Time 
        tp (sympy.Symbol): Time of peak amplitude
        tau (sympy.Symbol): Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Omega0 (sympy.Symbol): Initial Condition Frequency
        Omega_QNM (sympy.Symbol): Real part of Quasinormal mode (QNM) frequency (little omega)/(mode number)

    Returns:
        Omega - Strain frequency (waveform frequency)/(mode number)
    '''
    x = (t - tp)/tau
    return Omega_QNM * (Omega0 / Omega_QNM)**((sp.tanh(x) - 1) / sp.Integer(-2))

def BOB_strain_phase_sym(t, tp, Omega0, Omega_QNM, tau, Phi_0):
    '''

    Phase for the strain when the BOB amplitude models the strain (for t_0 = -inf)

    Args:
        t (sympy.Symbol): Time 
        tp (sympy.Symbol): Time of peak amplitude
        tau (sympy.Symbol): Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Omega0 (sympy.Symbol): Initial Condition Frequency
        Omega_QNM (sympy.Symbol): Real part of Quasinormal mode (QNM) frequency (little omega)/(mode number)
        Phi_0 (sympy.Symbol): Initial Condition Phase (phi)/(mode number)

    Returns:
        Phi -  Phase of strain (phi)/(mode number)
        
        Omega - Strain frequency (waveform frequency)/(mode number)
    '''
    Omega = BOB_strain_freq_sym(t, tp, Omega0, Omega_QNM, tau)
    x = (t - tp)/tau
    outer = tau / sp.Integer(2)
    log_sqrt_ratio = sp.Rational(1,2) * sp.log(Omega_QNM/Omega0)
    term1_arg = log_sqrt_ratio * (sp.tanh(x) + 1)
    term2_arg = log_sqrt_ratio * (sp.tanh(x) - 1)
    return outer * (Omega0 * Ei_sp(term1_arg) - Omega_QNM * Ei_sp(term2_arg)) + Phi_0, Omega

def BOB_news_freq_sym(t, tp, Omega0, Omega_QNM, tau):
    '''
    
    Frequency for the news when the BOB amplitude models the news (for t_0 = -inf)

    Args:
        t (sympy.Symbol): Time 
        tp (sympy.Symbol): Time of peak amplitude
        tau (sympy.Symbol): Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Omega0 (sympy.Symbol): Initial Condition Frequency
        Omega_QNM (sympy.Symbol): Real part of Quasinormal mode (QNM) frequency (little omega)/(mode number)

    Returns:
        Omega - News frequency (waveform frequency)/(mode number)
    '''

    x = (t - tp)/tau
    Omega_minus = Omega_QNM**2 - Omega0**2
    Omega_plus = Omega_QNM**2 + Omega0**2
    return sp.sqrt(Omega_minus * sp.tanh(x) / sp.Integer(2) + Omega_plus / sp.Integer(2))

def BOB_news_phase_sym(t, tp, Omega0, Omega_QNM, tau, Phi_0):
    '''
    Phase for the strain when the BOB amplitude models the strain (for t_0 = -inf)

    Args:
        t (sympy.Symbol): Time 
        tp (sympy.Symbol): Time of peak amplitude
        tau (sympy.Symbol): Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Omega0 (sympy.Symbol): Initial Condition Frequency
        Omega_QNM (sympy.Symbol): Real part of Quasinormal mode (QNM) frequency (little omega)/(mode number)
        Phi_0 (sympy.Symbol): Initial Condition Phase (phi)/(mode number)

    Returns:
        Phi - News phase (waveform phase)/(mode number)

        Omega - Strain frequency (waveform frequency)/(mode number)
        '''
    Omega = BOB_news_freq_sym(t, tp, Omega0, Omega_QNM, tau)
    outer = tau / sp.Integer(2)
    inner1 = sp.log((Omega + Omega_QNM) / sp.Abs(Omega - Omega_QNM))
    inner2 = sp.log((Omega + Omega0) / sp.Abs(Omega - Omega0))
    return outer * (Omega_QNM * inner1 - Omega0 * inner2) + Phi_0, Omega

def BOB_psi4_freq_sym(t, tp, Omega0, Omega_QNM, tau):
    '''
    Frequency for Psi4 when the BOB amplitude models Psi4 (for t_0 = -inf)

    Args:
        t (sympy.Symbol): Time 
        tp (sympy.Symbol): Time of peak amplitude
        tau (sympy.Symbol): Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Omega0 (sympy.Symbol): Initial Condition Frequency
        Omega_QNM (sympy.Symbol): Real part of Quasinormal mode (QNM) frequency (little omega)/(mode number)

    Returns:
        Omega - Psi4 frequency (waveform frequency) / (mode number)
    '''
    x = (t - tp)/tau
    k = (Omega_QNM**4 - Omega0**4) / sp.Integer(2)
    return (Omega0**4 + k * (sp.tanh(x) + 1))**sp.Rational(1, 4)

def BOB_psi4_phase_sym(t, tp, Omega0, Omega_QNM, tau, Phi_0):
    '''
    Phase for Psi4 when the BOB amplitude models Psi4 (for t_0 = -inf)

    Args:
        t (sympy.Symbol): Time 
        tp (sympy.Symbol): Time of peak amplitude
        tau (sympy.Symbol): Damping term; can also be described as 1/gamma (gamma is imaginry QNM fre)
        Omega0 (sympy.Symbol): Initial Condition Frequency
        Omega_QNM (sympy.Symbol): Real part of Quasinormal mode (QNM) frequency (little omega)/(mode number)
        Phi_0 (sympy.Symbol): Initial Condition Phase (phi)/(mode number)

    Returns:
        Phi - Psi4 Phase (waveform phase)/(mode number)

        Omega - Psi4 frequency (waveform frequency) / (mode number)
    '''
    Omega = BOB_psi4_freq_sym(t, tp, Omega0, Omega_QNM, tau)
    Omega_minus_q0 = Omega_QNM - Omega0
    Omega_plus_q0 = Omega_QNM + Omega0
    outer_num = sp.sqrt(Omega_minus_q0 * Omega_plus_q0) * tau
    outer_den = 2 * sp.sqrt(sp.Abs(Omega_minus_q0)) * sp.sqrt(sp.Abs(Omega_plus_q0))
    outer = outer_num / outer_den
    inner1 = Omega_QNM * (sp.log(sp.Abs(Omega + Omega_QNM)) - sp.log(sp.Abs(Omega - Omega_QNM)))
    inner2 = -Omega0 * (sp.log(sp.Abs(Omega + Omega0)) - sp.log(sp.Abs(Omega - Omega0)))
    inner3 = 2 * Omega_QNM * sp.atan(Omega / Omega_QNM)
    inner4 = -2 * Omega0 * sp.atan(Omega / Omega0)
    return outer * (inner1 + inner2 + inner3 + inner4) + Phi_0, Omega


# --- Amplitude ---
def BOB_amplitude(BOB):
    '''
    Returns the amplitude of the BOB waveform

    Args:
        BOB: BOB object

    Returns:
        A(t) - Waveform amplitude
    '''
    return BOB.Ap / np.cosh(BOB.t_tp_tau)

### --- FINITE t0 MODELS --- ###

def BOB_strain_freq_finite_t0(BOB):
    '''
    Frequency for strain for when the BOB amplitude models the strain (for finite t_0)

    Args:
        BOB: BOB object

    Returns:
        Omega - Strain frequency (waveform frequency)/(mode number)
    '''
    Omega_ratio = BOB.Omega_0/BOB.Omega_QNM
    tanh_t_tp_tau_m1 = np.tanh(BOB.t_tp_tau)-1
    tanh_t0_tp_tau_m1 = np.tanh(BOB.t0_tp_tau)-1
    return BOB.Omega_QNM*(Omega_ratio**(tanh_t_tp_tau_m1/tanh_t0_tp_tau_m1))

def BOB_strain_phase_finite_t0_numerically(BOB):
    '''
    Strain phase computed numerically for when the BOB amplitude models the strain (for finite t_0)

    Args:
        BOB: BOB object

    Returns:
        Phi - Phase of strain (phi)/(mode number)

        Omega - Strain frequency (waveform frequency)/(mode number)
    '''
    Omega = BOB_strain_freq_finite_t0(BOB)
    Phase = cumulative_trapezoid(Omega, BOB.t, initial=0)
    return Phase + BOB.Phi_0, Omega

def BOB_strain_phase_finite_t0(BOB):
    '''
    Strain phase for when the BOB amplitude models the strain (for finite t_0)

    Args:
        BOB: BOB object

    Returns:
        Phi - Strain phase (phi)/(mode number)

        Omega - Strain frequency (waveform frequency)/(mode number)
    '''
    try:
        Omega = BOB_strain_freq_finite_t0(BOB)
        outer = BOB.Omega_QNM*BOB.tau/2.
        Omega_ratio = BOB.Omega_0/BOB.Omega_QNM
        tanh_tp_t0_tau_p1 = np.tanh(-BOB.t0_tp_tau) + 1
        term1_exp = 2. / tanh_tp_t0_tau_p1
        term1 = Omega_ratio**term1_exp
        log_Omega_ratio_div_tanh = -np.log(Omega_ratio) / (np.tanh(BOB.t0_tp_tau) - 1)
        term2 = log_Omega_ratio_div_tanh * (np.tanh(BOB.t_tp_tau) + 1)
        term3 = log_Omega_ratio_div_tanh * (np.tanh(BOB.t_tp_tau) - 1)
        inner = term1 * Ei_np(term2) - Ei_np(term3)
        Phi = outer*inner + BOB.Phi_0
        return Phi, Omega
    except (ValueError, ZeroDivisionError):
        if BOB.auto_switch_to_numerical_integration:
            return BOB_strain_phase_finite_t0_numerically(BOB)
        else:
            raise ValueError("Analytical strain phase integration failed.")

def BOB_news_freq_finite_t0(BOB):
    '''
    Frequency for news for when the BOB amplitude models the news (for finite t_0)

    Args:
        BOB: BOB object

    Returns:
        Omega - News frequency (waveform frequency)/(mode number)
    '''
    F_denom = 1 - np.tanh(BOB.t0_tp_tau)
    if F_denom == 0: raise ValueError("Singularity: t0 -> +infinity is not a valid limit.")
    F = (BOB.Omega_QNM**2 - BOB.Omega_0**2) / F_denom
    Omega2 = BOB.Omega_QNM**2 + F * (np.tanh(BOB.t_tp_tau) - 1)
    if np.any(Omega2 < 0): raise ValueError("Imaginary frequency encountered. Check parameters.")
    return np.sqrt(Omega2)

def BOB_news_phase_finite_t0_numerically(BOB):
    '''
    News phase computed numerically for when the BOB amplitude models the news (for finite t_0)

    Args:
        BOB: BOB object

    Returns:
        Phi - News phase (phi)/(mode number)

        Omega - News frequency (waveform frequency)/(mode number)
    '''
    Omega = BOB_news_freq_finite_t0(BOB)
    Phase = cumulative_trapezoid(Omega, BOB.t, initial=0)
    return Phase + BOB.Phi_0, Omega

def BOB_news_phase_finite_t0(BOB):
    '''
    News phase for when the BOB amplitude models the news (for finite t_0)

    Args:
        BOB: BOB object

    Returns:
        Phi - News phase (phi)/(mode number)

        Omega - News frequency (waveform frequency)/(mode number)
    '''
    Omega = BOB_news_freq_finite_t0(BOB)
    F_denom = 1 - np.tanh(BOB.t0_tp_tau)
    F = (BOB.Omega_QNM**2 - BOB.Omega_0**2) / F_denom
    delta = 2 * F - BOB.Omega_QNM**2
    if delta <= 0:
        if BOB.auto_switch_to_numerical_integration:
            return BOB_news_phase_finite_t0_numerically(BOB)
        else:
            raise ValueError("Analytical news phase integration failed (delta <= 0).")
    term1_outer = BOB.Omega_QNM * BOB.tau / 2.
    term1_inner = np.log((Omega + BOB.Omega_QNM) / np.abs(Omega - BOB.Omega_QNM))
    term2_outer = -np.sqrt(delta) * BOB.tau
    term2_inner = np.arctan(Omega / np.sqrt(delta))
    Phi = term1_outer * term1_inner + term2_outer * term2_inner + BOB.Phi_0
    return Phi, Omega

def BOB_psi4_freq_finite_t0(BOB):
    '''
    Frequency for Psi4 for when the BOB amplitude models Psi4 (for finite t_0)

    Args:
        BOB: BOB object

    Returns:
        Omega - Psi4 frequency (waveform frequency) / (mode number)
    '''
    k_denom = 1 - np.tanh(BOB.t0_tp_tau)
    if k_denom == 0: raise ValueError("Singularity: t0 -> +infinity is not a valid limit.")
    k = (BOB.Omega_QNM**4 - BOB.Omega_0**4) / k_denom
    X = BOB.Omega_0**4 + k * (np.tanh(BOB.t_tp_tau) - np.tanh(BOB.t0_tp_tau))
    if np.any(X < 0): raise ValueError("Imaginary frequency encountered in Psi4. Check parameters.")
    return X**0.25

def BOB_psi4_phase_finite_t0_numerically(BOB):
    '''
    Psi4 phase computed numerically for when the BOB amplitude models Psi4 (for finite t_0)

    Args:
        BOB: BOB object

    Returns:
        Phi - Phase of Psi4 (phi)/(mode number)

        Omega - Psi4 frequency (waveform frequency) / (mode number)
    '''
    Omega = BOB_psi4_freq_finite_t0(BOB)
    Phase = cumulative_trapezoid(Omega, BOB.t, initial=0)
    return Phase + BOB.Phi_0, Omega

def BOB_psi4_phase_finite_t0(BOB):
    '''
    Psi4 phase for when the BOB amplitude models Psi4 (for finite t_0)

    Args:
        BOB: BOB object

    Returns:
        Phi - Phase of Psi4 (phi)/(mode number)

        Omega - Psi4 frequency (waveform frequency) / (mode number)
    '''
    try:
        Omega = BOB_psi4_freq_finite_t0(BOB)
        k = (BOB.Omega_QNM**4 - BOB.Omega_0**4)/(1-np.tanh(BOB.t0_tp_tau))
        KappaP_arg = BOB.Omega_0**4 + k*(1-np.tanh(BOB.t0_tp_tau))
        KappaM_arg = BOB.Omega_0**4 - k*(1+np.tanh(BOB.t0_tp_tau))
        if KappaP_arg < 0 or KappaM_arg < 0: raise ValueError("Invalid arguments for Kappa.")
        KappaP = KappaP_arg**0.25
        KappaM = KappaM_arg**0.25
        arctanhP = KappaP*BOB.tau*(0.5*np.log(((1+(Omega/KappaP))*(1-(BOB.Omega_0/KappaP)))/(((1-(Omega/KappaP)))*(1+(BOB.Omega_0/KappaP)))))
        arctanhM = KappaM*BOB.tau*(0.5*np.log(((1+(Omega/KappaM))*(1-(BOB.Omega_0/KappaM)))/(((1-(Omega/KappaM)))*(1+(BOB.Omega_0/KappaM)))))
        arctanP  = KappaP*BOB.tau*(np.arctan(Omega/KappaP) - np.arctan(BOB.Omega_0/KappaP))
        arctanM  = KappaM*BOB.tau*(np.arctan(Omega/KappaM) - np.arctan(BOB.Omega_0/KappaM))
        Phi = arctanhP + arctanP - arctanhM - arctanM
        return Phi + BOB.Phi_0, Omega
    except (ValueError, ZeroDivisionError, TypeError):
        if BOB.auto_switch_to_numerical_integration:
            return BOB_psi4_phase_finite_t0_numerically(BOB)
        else:
            raise ValueError("Analytical Psi4 (finite t0) integration failed.")

def BOB_strain_freq(BOB):
    '''
    Strain frequency for when the BOB amplitude models the strain (for t0 = -infinity)

    Args:
        BOB: BOB object

    Returns:
        Omega - Strain frequency (waveform frequency)/(mode number)
    '''
    
    Omega_ratio = BOB.Omega_0/BOB.Omega_QNM
    tanh_t_tp_tau_m1 = np.tanh(BOB.t_tp_tau)-1
    return BOB.Omega_QNM*(Omega_ratio**(tanh_t_tp_tau_m1/(-2.)))

def BOB_strain_phase(BOB):
    '''
    Strain phase for when the BOB amplitude models the strain (for t0 = -infinity)

    Args:
        BOB: BOB object

    Returns:
        Phi - Strain phase (phi)/(mode number)

        Omega - Strain frequency (waveform frequency)/(mode number)
    '''
    Omega = BOB_strain_freq(BOB)
    outer = BOB.tau/2.
    Omega_ratio = BOB.Omega_QNM/BOB.Omega_0
    log_sqrt_ratio = np.log(np.sqrt(Omega_ratio))
    term1 = log_sqrt_ratio*(np.tanh(BOB.t_tp_tau)+1)
    term2 = log_sqrt_ratio*(np.tanh(BOB.t_tp_tau)-1)
    inner  = BOB.Omega_0*Ei_np(term1) - BOB.Omega_QNM*Ei_np(term2)
    return outer*inner + BOB.Phi_0, Omega

def BOB_news_freq(BOB):
    '''
    News frequency for when the BOB amplitude models the news (for t0 = -infinity)

    Args:
        BOB: BOB object

    Returns:
        Omega - News frequency (waveform frequency)/(mode number)
    '''
    Omega_minus = BOB.Omega_QNM**2 - BOB.Omega_0**2
    Omega_plus  = BOB.Omega_QNM**2 + BOB.Omega_0**2
    Omega2 = Omega_minus*np.tanh(BOB.t_tp_tau)/2. + Omega_plus/2.
    if np.any(Omega2 < 0): raise ValueError("Imaginary frequency.")
    return np.sqrt(Omega2)

def BOB_news_phase(BOB):
    '''
    News phase for when the BOB amplitude models the news (for t0 = -infinity)

    Args:
        BOB: BOB object

    Returns:
        Phi - News phase (phi)/(mode number)

        Omega - News frequency (waveform frequency)/(mode number)
    '''
    if(BOB.Omega_0==0): raise ValueError("Omega_0 cannot be zero")
    Omega = BOB_news_freq(BOB)
    Omega_minus_Q = np.abs(Omega - BOB.Omega_QNM)
    Omega_minus_0 = np.abs(Omega - BOB.Omega_0)
    outer = BOB.tau/2.
    inner1 = np.log(Omega + BOB.Omega_QNM) - np.log(Omega_minus_Q)
    inner2 = np.log(Omega + BOB.Omega_0) - np.log(Omega_minus_0)
    return outer*(BOB.Omega_QNM*inner1 - BOB.Omega_0*inner2) + BOB.Phi_0, Omega

def BOB_psi4_freq(BOB):
    '''
    Psi4 frequency for when the BOB amplitude models the Psi4 (for t0 = -infinity)

    Args:
        BOB: BOB object

    Returns:
        Omega - Psi4 frequency (waveform frequency) / (mode number)
    '''
    k = (BOB.Omega_QNM**4 - BOB.Omega_0**4)/2.
    X = BOB.Omega_0**4 + k*(np.tanh(BOB.t_tp_tau) + 1)
    if np.any(X < 0): raise ValueError("Imaginary frequency.")
    return X**0.25

def BOB_psi4_phase(BOB):        
    '''
    Psi4 phase for when the BOB amplitude models the Psi4 (for t0 = -infinity)

    Args:
        BOB: BOB object

    Returns:
        Phi - Phase of Psi4 (phi)/(mode number)

        Omega - Psi4 frequency (waveform frequency) / (mode number)
    '''
    Omega = BOB_psi4_freq(BOB)
    Omega_minus_q0 = BOB.Omega_QNM - BOB.Omega_0
    Omega_plus_q0  = BOB.Omega_QNM + BOB.Omega_0
    outer_num = np.sqrt(Omega_minus_q0*Omega_plus_q0)*BOB.tau
    outer_den = 2*np.sqrt(np.abs(Omega_minus_q0))*np.sqrt(np.abs(Omega_plus_q0))
    outer = outer_num/outer_den
    inner1 = BOB.Omega_QNM*(np.log(np.abs(Omega+BOB.Omega_QNM)) - np.log(np.abs(Omega-BOB.Omega_QNM)))
    inner2 = -BOB.Omega_0 * (np.log(np.abs(Omega+BOB.Omega_0)) - np.log(np.abs(Omega-BOB.Omega_0)))
    inner3 = 2*BOB.Omega_QNM*np.arctan(Omega/BOB.Omega_QNM)
    inner4 = -2*BOB.Omega_0*np.arctan(Omega/BOB.Omega_0)
    return outer*(inner1+inner2+inner3+inner4) + BOB.Phi_0, Omega

